import sys,os
import numpy as np
import matplotlib.pyplot as plt
import numpy.fft as fft
import pywt
import math
import scipy.misc
from scipy.stats import norm
from multiprocessing import Process, cpu_count, Manager
from genSamples import getKspace,getVenc
from CSRecoverySuite import CSRecovery, CSRecoveryDebiasing, Operator4dFlow, pywt2array, array2pywt, crop
# Import Solvers
from maps import OperatorWaveletToFourier
from solver_l1_norm import RecoveryL1NormNoisy, MinimizeSumOfSquares
from solver_omp import OMPRecovery

import argparse

home = os.getenv('HOME')

CS_MODE     = 0
DEBIAS_MODE = 1
OMP_MODE    = 2

# OMP algorithm defaults
def_omp_mode     = 'stomp'
def_omp_iter     = 10
def_omp_ts       = 2.0
def_omp_usenorms = False

def get_eta(im, imNrm, noise_percent, m):
  avgnorm = imNrm/math.sqrt(im.size)
  stdev   = noise_percent * avgnorm
  rv      = norm()
  # ppf: inverse of cdf
  eta     = stdev*math.sqrt(2*m + 2*math.sqrt(m)*rv.ppf(0.95))
  if eta < 1E-3: # min. threshold for eta
    eta = 1E-3
  return eta

def recoverOne(kspace, imsz, eta, omega, wvlt='haar', 
               solver_mode=CS_MODE, omp_mode=def_omp_mode, omp_iter=def_omp_iter, omp_ts=def_omp_ts, omp_usenorms=def_omp_usenorms):
  
  A = OperatorWaveletToFourier(imsz, samplingSet=omega, waveletName=wvlt)
  
  wim = pywt.wavedec2(fft.ifft2(kspace, norm='ortho'), wavelet=wvlt, mode='zero')
  wim = pywt.coeffs_to_array(wim)[0]

  yim = A.eval(wim, 1)
  
  # each recovery method returns (solution, norm) 
  # here we're only interested in the solution
  if(solver_mode == OMP_MODE):
    
    # OMP Recovery
    tol = eta/np.linalg.norm(yim.ravel(),2)
    print('Recovering using OMP with tol = %8.3e' % (tol))

    wim = OMPRecovery(A, yim, tol=tol, ompMethod=omp_mode, maxItns=omp_iter, ts_factor=omp_ts, useNorms=omp_usenorms)[0]

  else:
    # CS Recovery
    print('Recovering using CS with eta =', eta)
    wim =  RecoveryL1NormNoisy(eta, yim, A, disp=True)

  if isinstance(wim, tuple):
    wim = wim[0] # for the case where ynrm is less than eta
  
  # CS-DEBIAS Recovery
  if(solver_mode == DEBIAS_MODE):
    support = (np.abs(wim.flatten()) > 1E-3)
    Adeb = A.colRestrict(basisSet=support)
    wim =  MinimizeSumOfSquares(yim, Adeb)
    if isinstance(wim, tuple):
      wim = wim[0] # for the case where ynrm is less than eta

  # Reconstruct Image from wavelet coefficients
  csim = A.getImageFromWavelet(wim)

  return csim
  
def recover(noisy, original, pattern, noise_percent, processnum, return_dict, wvlt='haar', 
            solver_mode=CS_MODE, omp_mode=def_omp_mode, omp_iter=def_omp_iter, omp_ts=def_omp_ts, omp_usenorms=def_omp_usenorms):    
  imsz = crop(noisy[0,0,0]).shape
  cs = np.zeros(noisy.shape[0:3] + imsz,dtype=complex)
  print('Pattern shape: ', pattern.shape)
  for n in range(noisy.shape[0]):
    if len(pattern.shape) > 2:
      if pattern.shape[0] == 1:
        idx = 0
      else:
        idx = n
      omega = crop(pattern[idx])
    else:
      omega = crop(pattern)
    for k in range(noisy.shape[1]):
      for j in range(noisy.shape[2]):
        im = crop(original[0,k,j])
        imNrm=np.linalg.norm(im.ravel(), 2)
        eta = get_eta(im, imNrm, noise_percent, imsz[0])
        cs[n,k,j] = recoverOne(crop(noisy[n,k,j]), imsz, eta, omega, wvlt, solver_mode, omp_mode, omp_iter, omp_ts,omp_usenorms)
        print('Recovered! Repetition: %d, Image: %d, Component: %d' % (n,k,j))

  return_dict[processnum] = cs
  return cs

def recoverAll(fourier_file, orig_file, pattern, noise_percent, c=2, wvlt='haar', 
               mode=CS_MODE, omp_mode=def_omp_mode, omp_iter=def_omp_iter, omp_ts=def_omp_ts, omp_usenorms=def_omp_usenorms):
	
  # Load data
  if isinstance(fourier_file,str):
    data = np.load(fourier_file)
  else:
    data = fourier_file
  
  # Load original image
  if isinstance(orig_file,str):
    original = np.load(orig_file)
  else:
    original = orig_file

  shp = data.shape[0:3] + crop(np.zeros(data.shape[3:])).shape
  recovered = np.empty(shp,dtype=np.complex64)
  interval = max(int(data.shape[0]/c),1)
  manager = Manager()
  return_dict = manager.dict()
  jobs = []
  imsz = crop(original[0,0,0]).shape
  first = original[0,0,0]  
  for n in range(0, data.shape[0], interval):
    if pattern.shape[0] > 1:
      pattern_sample = pattern[n:(n+interval)]
    else:
      pattern_sample = pattern
    p = Process(target=recover, args=(data[n:n+interval], original, pattern_sample, noise_percent, int(n/interval), return_dict, 
                                      wvlt, mode, omp_mode, omp_iter, omp_ts, omp_usenorms))
    jobs.append(p)
    p.start()
  for job in jobs:
    job.join()
  recovered = np.concatenate([v for k,v in sorted(return_dict.items())], axis=0)
  print('Finished recovering, with final shape', recovered.shape)
  return recovered

def recover_vel(compleximg, venc=None, threshold=True):
  vel = np.zeros(compleximg.shape)
  for n in range(0,compleximg.shape[0]):
    for j in range(0, compleximg.shape[2]):
      m = compleximg[n,0,j]
      vel[n,0,j] = np.abs(m)
      for k in range(1,compleximg.shape[1]):
        v = compleximg[n,k,j]
        # Formula in the paper
        v = venc/(np.pi)*(np.angle(v) - np.angle(m))
        vel[n,k,j] = v
  return vel

def phase_info(compleximg):
  #return reference phase and magnitude of velocities
  #required for inversion from velocity components
  refphase = np.angle(compleximg[:,0,:])
  return refphase, np.abs(compleximg)

def linear_reconstruction(fourier_file, omega=None):
  # Check if fourier_file is file or string
  if isinstance(fourier_file, str): 
    kspace = np.load(fourier_file)
  else:
    kspace = fourier_file

  # Check Omega if 2D or 3D with multiple undersampling patterns
  if omega is not None:
    # Check the size of the 3D mask, either one or equal to the samples
    if(len(omega.shape) == 3):
      if((omega.shape[0] != 1)and(omega.shape[0] != kspace.shape[0])):
        print('Invalid number of samples in 3D mask.')
        exit(-1)
    # Multiple mask case
    if(len(omega.shape) == 3):
      for loopA in range(omega.shape[0]):      
        omega[loopA] = crop(omega[loopA])
    elif(len(omega.shape) == 2):
      omega = crop(omega)
    else:
      print("ERROR: mask has unexpected shape.")
      sys.exit(-1)

  # Perform linear reconstructions      
  imsz = crop(kspace[0,0,0]).shape
  kspace = kspace[:,:,:, :imsz[0], :imsz[1]]
  linrec = np.zeros(kspace.shape[0:3] + imsz, dtype=complex)
  for n in range(kspace.shape[0]):
    for k in range(kspace.shape[1]):
      for j in range(kspace.shape[2]):
        if omega is not None:
          if(len(omega.shape) == 2):
            kspace[n,k,j][~omega] = 0
          elif((len(omega.shape) == 3)and(omega.shape[0] == 1)):
            kspace[n,k,j][~omega[0]] = 0
          else:
            kspace[n,k,j][~omega[n]] = 0
        linrec[n,k,j] = fft.ifft2(crop(kspace[n,k,j]), norm='ortho')
  return linrec

def getMethodString(mtd):
  if(mtd == 0):
    return 'CS'
  elif(mtd == 1):
    return 'CSDEB'
  elif(mtd == 2):
    return 'OMP'

# Get File Names
def getFiles(args):
  
  # Noisy Fourier measurement file
  fourier_file = args.fromdir + 'noisy_noise' + str(int(args.noisepercent*100)) + '_n' + str(args.repetitions) + '.npy'

  # Undersampling mask file
  if not args.usemultipatterns: 
    mask_file = args.maskdir + 'undersamplpattern_p' + str(int(args.uVal*100)) + args.uType + '.npy'
  else:
    mask_file = args.maskdir + 'undersamplpattern_p' + str(int(args.uVal*100)) + args.uType + '_n' + str(args.repetitions) + '.npy'
  
  # Image file
  orig_file = args.fromdir + 'imgs_n1.npy'

  # Reconstruction file
  rfs = str(int(args.noisepercent*100)) + '_p' + str(int(args.uVal*100)) + args.uType + '_n' + str(args.repetitions) + '_w' + args.wavelet.upper() + '_a' + getMethodString(args.method) + '.npy' 

  rec_file     = args.recdir + 'rec_noise' + rfs  
  # Save linear reconstruction file and velocities if requested
  rec_lin_file = None
  vel_file     = None
  vel_lin_file = None
  if(args.evallinrec):
    rec_lin_file = args.recdir + 'lin_noise' + rfs    
  if(args.savevels):
    vel_file = args.recdir + 'rec_vel_noise' + rfs  
    if(args.evallinrec):
      vel_lin_file = args.recdir + 'lin_vel_noise' + rfs  
  
  # Return file names
  return fourier_file,mask_file,orig_file,rec_file,vel_file,rec_lin_file,vel_lin_file

# MAIN 
if __name__ == '__main__':

  parser = argparse.ArgumentParser(description='.')

  # noisePercent    = 0.0
  parser.add_argument('-n', '--noisepercent',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=0.0,
                      type=float,
                      choices=None,
                      required=False,
                      help='noise percent based the average two-norm of the k-space image',
                      metavar='',
                      dest='noisepercent')
    
  # uVal            = 0.75,
  parser.add_argument('-v', '--urate',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=0.75,
                      type=float,
                      choices=None,
                      required=False,
                      help='undersampling rate',
                      metavar='',
                      dest='uVal')

  # uType           = 'vardengauss',
  parser.add_argument('-u', '--utype',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='vardengauss',
                      type=str,
                      choices=['bernoulli','vardentri','vardengauss','vardenexp','halton'],
                      required=False,
                      help='undersampling pattern type',
                      metavar='',
                      dest='uType')

  # numRealizations = 1
  parser.add_argument('-r', '--repetitions',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=1,
                      type=int,
                      choices=None,
                      required=False,
                      help='number of k-space samples and mask to generate',
                      metavar='',
                      dest='repetitions')

  # numProcesses = 2
  parser.add_argument('-c', '--numprocesses',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=2,
                      type=int,
                      choices=None,
                      required=False,
                      help='number of threads for parallel image reconstruction',
                      metavar='',
                      dest='numprocesses')

  # fromdir         = '\.'
  parser.add_argument('-f', '--fromdir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='folder for original image, k-space images and velocity encoding',
                      metavar='',
                      dest='fromdir')

  # recdir         = '\.'
  parser.add_argument('-d', '--recdir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='folder for recontructed images',
                      metavar='',
                      dest='recdir')

  # maskdir         = '\.'
  parser.add_argument('-m', '--maskdir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='folder for undersampling patterns',
                      metavar='',
                      dest='maskdir')

  # method
  parser.add_argument('-rm', '--method',
                      action=None,
                      # nargs='*',
                      const=None,
                      default=0,
                      type=int,
                      choices=[0,1,2],
                      required=False,
                      help='nonlinear reconstruction method',
                      metavar='',
                      dest='method')

  # Type of OMP solver
  parser.add_argument('--ompmode',
                      action=None,
                      # nargs='*',
                      const=None,
                      default='stomp',
                      type=str,
                      choices=['stomp','omp'],
                      required=False,
                      help='Greedy reconstruction method, either omp or stomp',
                      metavar='',
                      dest='ompmode')

  # omp threshold factor
  parser.add_argument('--ompts',
                      action=None,
                      const=None,
                      default=2.0,
                      type=float,
                      choices=None,
                      required=False,
                      help='threshold factor in stomp',
                      metavar='',
                      dest='ompts')
  
  # Maximum number of iterations for OMP/STOMP
  parser.add_argument('--ompiter',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=10,
                      type=int,
                      choices=None,
                      required=False,
                      help='maximum number of iterations for OMP/STOMP',
                      metavar='',
                      dest='ompiter')

    # save velocities
  parser.add_argument('--ompusenorms',
                    action='store_true',
                    default=False,
                    required=False,
                    help='Include column norms in OMP mathing phase',
                    dest='ompusenorms')    

  # wavelet type
  parser.add_argument('-w', '--wavelet',
                      action=None,
                      # nargs='*',
                      const=None,
                      default='haar',
                      type=str,
                      choices=['haar','db4','db8'],
                      required=False,
                      help='wavelet family',
                      metavar='',
                      dest='wavelet')

  # compute linear reconstruction
  parser.add_argument('--evallinrec',
                      action='store_true',
                      default=False,
                      required=False,
                      help='Evaluate linear reconstructions',
                      dest='evallinrec')    
    
  # use unique undesampling pattern for each reconstruction (false) or change every time
  parser.add_argument('-um', '--usemultipatterns',
                      action='store_true',
                      default=False,
                      required=False,
                      help='generate a unique undersampling pattern for each noise realization',
                      dest='usemultipatterns')

  # save velocities
  parser.add_argument('--savevels',
                    action='store_true',
                    default=False,
                    required=False,
                    help='Save velocity fields',
                    dest='savevels')    

  # Parse Commandline Arguments
  args = parser.parse_args()
  
  # Get File Names
  fourier_file,mask_file,orig_file,rec_file,vel_file,rec_lin_file,vel_lin_file = getFiles(args)
  
  # Perform Reconstruction
  if not os.path.exists(rec_file):

    print('Loading undersampling mask...')    
    if os.path.exists(mask_file):
      umask = np.load(mask_file)
    else:
      print('ERROR: Undersampling mask file not found: ',mask_file)
      sys.exit(-1)

    print('Computing reconstructions...')
    recovered = recoverAll(fourier_file, 
                           orig_file, 
                           umask, 
                           args.noisepercent, 
                           c=args.numprocesses, 
                           wvlt=args.wavelet, 
                           mode=args.method,
                           omp_mode=args.ompmode, 
                           omp_iter=args.ompiter, 
                           omp_ts=args.ompts,
                           omp_usenorms=args.ompusenorms)
    print('Saving reconstruction to file: ',rec_file)
    np.save(rec_file, recovered)
  else:
    print('Retrieving recovered images from numpy file: ',rec_file)
    recovered = np.load(rec_file)

  # Linear Reconstructions
  if(args.evallinrec):
    print('Computing linear reconstrucitons...')
    linrec = linear_reconstruction(fourier_file, umask)
    print('Saving linear reconstrucitons to file: ',rec_lin_file)
    np.save(rec_lin_file, linrec)

  # Save velocities
  if(args.savevels): 

    # Load velocity encoding
    print('Loading velocity encoding...')
    vencfile = args.fromdir + 'venc_n1.npy'
    if os.path.exists(vencfile):
      venc = np.load(vencfile)
    else:
      print('WARNING: file for velocity encoding not found: ',vencfile)
      venc = None  

    print('Computing velocities...')  
    recovered = recover_vel(recovered, venc)
    print('Saving velocities to file: ',vel_file) 
    np.save(vel_file, recovered)
    
    if(args.evallinrec):
      linrec = recover_vel(linrec, venc)
      print('Saving linear reconstructed velocities to file: ',vel_lin_file)
      np.save(vel_lin_file, linrec)



# %% 
import numpy as np
import numpy.fft as fft
import numpy.linalg as la
import pywt

# Generic Operator
class genericOperator(object):
    pass

# Implementation of a simple linear operator
class OperatorLinear(genericOperator):
  def __init__(self, mat, samplingSet=None, basisSet=None):
    self.__mat = mat
    self.__shape = mat.shape

    # Define input and output shape for post-multiplication
    self.inShape = self.__mat.shape[1]
    self.outShape = self.__mat.shape[0]

    # Define input and output shape for post-multiplication
    self.wavShape = self.__mat.shape[1]
    self.imShape = self.__mat.shape[0]

    # Assign sampling set
    if isinstance(samplingSet, list):
      samplingSet = np.array(samplingSet)
    tmp = np.zeros(self.outShape,dtype=bool)
    tmp[samplingSet] = True
    self.samplingSet = tmp

    # Assign basis set
    if isinstance(basisSet, list):
      basisSet = np.array(basisSet)
    tmp = np.zeros(self.inShape,dtype=bool)
    tmp[basisSet] = True
    self.basisSet = tmp

  def eval(self, x, mode=1):
    if(mode==1):
      
      # Direct map

      if(self.basisSet is not None):
        tmp = np.dot(self.__mat[:,self.basisSet],x[self.basisSet])
      else: 
        tmp = np.dot(self.__mat,x)

      if(self.samplingSet is not None):
        return tmp[self.samplingSet]
      else:
        return tmp

    elif(mode==2):
      
      # Adjoint map

      if(self.samplingSet is not None):
        tmp = np.dot(np.conjugate(self.__mat.T[:,self.samplingSet]),x[self.samplingSet])
      else:
        tmp = np.dot(np.conjugate(self.__mat.T),x)

      if(self.basisSet is not None):
        return tmp[self.basisSet]
      else: 
        return tmp

  def adjoint(self, x):
      return self.eval(x, mode=2)

  @property
  def shape(self):
    "The shape of the operator."
    return self.__shape

  def input_size(self):
    return self.__shape[1]

  @property
  def T(self):
    "Transposed of the operator."
    return OperatorLinear(np.conjugate(self.__mat.T))

  def __matmul__(self, x):
    return np.dot(self.__mat,x)

  def colRestrict(self,idxSet=None):
    return OperatorLinear(self.__mat[:,idxSet])

  def norm(self):
    return np.linalg.norm(self.__mat,2)

# %%
# Find the name of the wavelet associated to the adjoint of
# the wavelet transform
def getAdjointWavelet(waveletName):
    if waveletName is None or waveletName == 'None':
        return None
    if len(waveletName) > 4 and waveletName[0:4] == 'bior':
        return 'rbio' + waveletName[4::]
    if len(waveletName) > 4 and waveletName[0:4] == 'rbio':
        return 'bior' + waveletName[4::]
    return waveletName

def getWaveletTransformShape(imShape, waveletName, waveletLevel=None):
    if waveletName is None or waveletName == 'None':
        return imShape

    x = np.zeros(imShape)
    Wx = pywt.wavedec2(x, waveletName, mode='zero', level=waveletLevel)

    return pywt.coeffs_to_array(Wx)[0].shape

def getWaveletTransformSlices(imShape, waveletName, waveletLevel=None):
    if waveletName is None or waveletName == 'None':
        return None

    x = np.zeros(imShape)
    Wx = pywt.wavedec2(x, waveletName, mode='zero', level=waveletLevel)

    return pywt.coeffs_to_array(Wx)[1]

def getWaveletReconstructionShape(imShape, waveletName, waveletLevel=None):
    if waveletName is None or waveletName == 'None':
        return imShape

    Wx = pywt.wavedec2(np.zeros(imShape), waveletName, mode='zero', level=waveletLevel)

    return pywt.waverec2(Wx, wavelet=waveletName, mode='zero').shape

# OperatorWaveletToFourier
#   This implements the linear map that takes as inputs
#   the wavelet coefficients of a complex image, and 
#   outputs the Fourier coefficients of the image. The 
#   map supports restricting the support of the wavelet
#   coefficients, and subsampling the Fourier coefficients
class OperatorWaveletToFourier(genericOperator):
    def __init__(self, imShape, samplingSet=None, basisSet=None, isTransposed=False, waveletName=None, waveletLevel=None):
        # Check for boundary case
        if waveletName == 'None':
            waveletName = None
        # Parameters
        self.imShape = imShape
        self.waveletName = waveletName
        self.waveletNameAdj = getAdjointWavelet(waveletName)
        self.waveletLevel = waveletLevel
        self.wavShape = getWaveletTransformShape(imShape, waveletName, waveletLevel)
        self.wavSlices = getWaveletTransformSlices(imShape, waveletName, waveletLevel)
        self.isTransposed = isTransposed
        self._norm = None
        # Test if the image shape produces a consistent reconstruction
        # using pyWavelets or not
        xrecShape = getWaveletReconstructionShape(imShape, waveletName, waveletLevel)
        self.waveletCrop = [ xrecShape[0] != imShape[0], 
                             xrecShape[1] != imShape[1] ]
        # Validate shapes
        if samplingSet is not None:

            # The sampling set should represent a 2D complex image
            if isinstance(samplingSet, list):
                samplingSet = np.array(samplingSet)
            if(samplingSet.ndim < 2):
                tmp = np.zeros(self.imShape,dtype=bool).flatten()
                tmp[samplingSet] = True
                samplingSet = tmp.reshape(self.imShape)
            # Check if the size is correct
            if samplingSet.shape[0] != self.imShape[0] or samplingSet.shape[1] != self.imShape[1]:
                raise ValueError('The sampling array does not match the shape of the image.')

        if basisSet is not None:

            # If the basisSet is given in terms of an indexset and not a binary mask, convert it to a binary mask
            # Basis Set should refer to a 2D Wavelet coefficient representation
            if isinstance(basisSet, list):
                basisSet = np.array(basisSet)
            if(basisSet.ndim < 2):
                tmp = np.zeros(self.wavShape,dtype=bool).flatten()
                tmp[basisSet] = True
                basisSet = tmp.reshape(self.wavShape)
            # Check if the size is correct
            if basisSet.shape[0] != self.wavShape[0] or basisSet.shape[1] != self.wavShape[1]:
                raise ValueError('The basis indices do not match the shape of the wavelet transform.') 

        # Restriction of the support of the wavelet coefficients
        self.basisSet = basisSet
        if basisSet is None:
            basisShape = self.wavShape
        else:
            basisShape = (np.count_nonzero(basisSet),)
        # Subsampling of Fourier coefficients
        self.samplingSet = samplingSet
        if samplingSet is None:
            samplingShape = self.imShape
        else:
            samplingShape = (np.count_nonzero(samplingSet),)
        # Input and output shapes
        if isTransposed:
            self.inShape = samplingShape
            self.outShape = basisShape
        else:
            self.inShape = basisShape
            self.outShape = samplingShape

    # The method eval is the only one that should use the self.isTransposed flag.
    def eval(self, x, mode=1):
        # print("shape of x: ",x.shape)
        # Verify if the instance is transposed
        if self.isTransposed:
            if mode == 1:
                mode = 2
            else:
                mode = 1
        # Evaluate forward map
        if mode == 1:

            # Check input dimension
            if(x.shape != self.wavShape):
              raise ValueError('ERROR: Input for direct application of the operator has not the correct size.')

            # Verify if the support of the wavelet transform is restricted
            if self.waveletName is None:
                if self.basisSet is None:
                    _x = x
                else:
                    _x = np.zeros(self.wavShape, dtype=np.complex)
                    _x[self.basisSet] = x[self.basisSet]
            else:
                if self.basisSet is None:
                    _w = pywt.array_to_coeffs(x, self.wavSlices, output_format='wavedec2')
                else:
                    # Remove the wavelet coefficients not in the basis set
                    x[np.logical_not(self.basisSet)] = 0.0                    
                    _w = pywt.array_to_coeffs(x, self.wavSlices, output_format='wavedec2')
                # Compute image from wavelet coefficients
                _x = pywt.waverec2(_w, wavelet=self.waveletName, mode='zero')
            # Verify if reconstruction is consistent
            if self.waveletCrop[0]:
                _x = _x[0:self.imShape[0], :]
            if self.waveletCrop[1]:
                _x = _x[:, 0:self.imShape[1]]
            # Verify if there is subsampling of the Fourier coefficients
            if self.samplingSet is None:
                return fft.fft2(_x, norm='ortho')
            else:
                _f = fft.fft2(_x, norm='ortho')
                _f[np.logical_not(self.samplingSet)] = 0.0 # Set the Fourier coefficients outside the set to zero
                # return _f[self.samplingSet]
                return _f

        # Evaluate adjoint
        if mode == 2:

            # Check input dimension
            if(x.shape != self.imShape):
              raise ValueError('ERROR: Input for inverse application of the operator has not the correct size.')

            # Verify if there is subsampling of the Fourier coefficients
            if self.samplingSet is None:
                _im = fft.ifft2(x, norm='ortho')
            else:
                _f = np.zeros(self.imShape, dtype=np.complex)
                _f[self.samplingSet] = x[self.samplingSet]
                # _f = x
                # _f[np.logical_not(self.samplingSet)] = 0.0
                _im = fft.ifft2(_f, norm='ortho')
            if self.waveletName is None:
                _w = _im
            else:
                _w = pywt.wavedec2(_im, wavelet=self.waveletNameAdj, mode='zero', level=self.waveletLevel)
                _w = pywt.coeffs_to_array(_w)[0]
            # Verify if the support of the wavelet transform is restricted
            if self.basisSet is None:
                return _w
            else:
                _w[np.logical_not(self.basisSet)] = 0.0
                # return _w[self.basisSet]
                return _w

    def adjoint(self, x):
        return self.eval(x, mode=2)

    def norm(self, maxItns=1E3, absTol=1E-6, relTol=1E-9):
        if self._norm is not None:
            return self._norm
        # Initialize variables
        x = np.random.normal(size=self.wavShape) + 1j * np.random.normal(size=self.wavShape)
        x = x / la.norm(x)
        s = 0
        ds = np.inf
        itn = 0
        stop = False
        # Power iteration loop
        while not stop:
            # Evaluate xp = A'A x
            xp = self.adjoint(self.eval(x))
            # Evaluate x' A'Ax to estimate sigma_max^2
            sp = np.sqrt(np.real(np.sum(np.conj(x.flatten()) * xp.flatten())))
            ds = np.abs(sp - s)
            if ds < absTol or ds < absTol * s or itn > maxItns:
                stop = True
            # Normalize singular vector
            x  = xp / la.norm(xp)
            s  = sp
        return s

    def getImageFromWavelet(self, x):
        if(x.shape != self.wavShape):
            raise ValueError('ERROR: Invalid input size for getImageFromWavelet.')
        if self.waveletName is None:
            if self.basisSet is None:
                _x = x
            else:
                _x = np.zeros(self.imShape, dtype=np.complex)
                _x[self.basisSet] = x[:]
        else:
            if self.basisSet is None:
                _w = pywt.array_to_coeffs(x, self.wavSlices, output_format='wavedec2')
            else:
                # Set to zero wavelet coefficients not in the basis set 
                x[np.logical_not(self.basisSet)] = 0.0
                _w = pywt.array_to_coeffs(x, self.wavSlices, output_format='wavedec2')
            # Compute image from wavelet coefficients
            _x = pywt.waverec2(_w, wavelet=self.waveletName, mode='zero')
        # Verify if reconstruction is consistent
        if self.waveletCrop[0]:
            _x = _x[0:self.imShape[0], :]
        if self.waveletCrop[1]:
            _x = _x[:, 0:self.imShape[1]]
        return _x

    def getImageFromFourier(self, y):
        if self.samplingSet is None:
            _im = fft.ifft2(y, norm='ortho')
        else:
            y[np.logical_not(self.samplingSet)] = 0.0
            _im = fft.ifft2(y, norm='ortho')
        return _im

    @property
    def shape(self):
        # This is the shape of the matrix representing
        # the operator with vectorized inputs and outputs
        return (np.prod(self.outShape), np.prod(self.inShape))

    def __matmul__(self, x):
        _y = self.eval(np.reshape(x, newshape=self.inShape))
        return _y.ravel()
        
    @property
    def T(self):
        # Instantiate the transpose of the operator
        return OperatorWaveletToFourier(imShape=self.imShape, samplingSet=self.samplingSet, basisSet=self.basisSet, isTransposed=not(self.isTransposed), waveletName=self.waveletName, waveletLevel=self.waveletLevel)

    # Create An operator with a 
    def colRestrict(self, basisSet=None):
        # Instantiate operator restricted to some entries
        return OperatorWaveletToFourier(imShape=self.imShape, samplingSet=self.samplingSet, basisSet=basisSet, isTransposed=self.isTransposed, waveletName=self.waveletName, waveletLevel=self.waveletLevel)

    # Print Operator
    def __str__(self):
        res  = '--- Wavelet to Fourier Operator\n'
        res += 'Shape of the original image: %d x %d\n' % (self.imShape[0],self.imShape[1])
        res += 'Wavelet name: %s\n' % (self.waveletName)
        res += 'Wavelet adjoint name: %s\n' % (self.waveletNameAdj)
        if self.waveletLevel is not None:
            res += 'Wavelet level: %d\n' % (self.waveletLevel)
        else:
            res += 'Wavelet level is not defined\n'
        res += 'Shape of the wavelet transform: %d x %d\n' % (self.wavShape[0],self.wavShape[1])
        res += 'Is operator transposed: ' + str(self.isTransposed) + '\n'
        # Sampling and basis sets
        if self.samplingSet is not None:
            res += 'Sampling set has size: %d x %d\n' % (self.samplingSet.shape[0],self.samplingSet.shape[1])
        else:
            res += 'Sampling set is not defined\n'
        if self.basisSet is not None:
            res += 'Basis set has size: %d x %d\n' % (self.basisSet.shape[0],self.basisSet.shape[1])
        else:
            res += 'Basis set is not defined\n'
        # Input and output shapes
        res += 'Input shape is: ' + str(self.inShape) + '\n'
        res += 'Output shape is: ' + str(self.outShape) + '\n'
        return res

# OperatorWaveletToFourierX4
#   This implements the linear map that takes as inputs
#   the wavelet coefficients of 4 complex images, and 
#   outputs their Fourier coefficients. The map supports restricting 
#   the support of the wavelet coefficients, and subsampling
#   the Fourier coefficients
class OperatorWaveletToFourierX4(genericOperator):
    def __init__(self, imShape, samplingSet=None, basisSet=None, isTransposed=False, waveletName='haar', waveletLevel=None):
        # Parameters
        self.imShape = imShape
        self.waveletName = waveletName
        self.waveletNameAdj = getAdjointWavelet(waveletName)
        self.waveletLevel = waveletLevel
        self.wavShape = getWaveletTransformShape(imShape, waveletName, waveletLevel)
        self.wavSlices = getWaveletTransformSlices(imShape, waveletName, waveletLevel)
        self.isTransposed = isTransposed
        self.samplingSet = samplingSet
        self.basisSet = basisSet
        self._norm = None
        # Generate array of maps
        if samplingSet is None and basisSet is None:
            self.map = [ OperatorWaveletToFourier(imShape=imShape, samplingSet=None, basisSet=None, isTransposed=isTransposed, waveletName=waveletName, waveletLevel=waveletLevel) for I in range(4) ]
        else:
            if samplingSet is None:
                if basisSet.ndim == 3:
                    self.map = [ OperatorWaveletToFourier(imShape=imShape, samplingSet=None, basisSet=basisSet[:, :, I], isTransposed=isTransposed, waveletName=waveletName, waveletLevel=waveletLevel) for I in range(4) ]
                else:
                    self.map = [ OperatorWaveletToFourier(imShape=imShape, samplingSet=None, basisSet=basisSet, isTransposed=isTransposed, waveletName=waveletName, waveletLevel=waveletLevel) for I in range(4) ]
            else:
                if samplingSet.ndim == 3:
                    if basisSet is None:
                        self.map = [ OperatorWaveletToFourier(imShape=imShape, samplingSet=samplingSet[:, :, I], basisSet=None, isTransposed=isTransposed, waveletName=waveletName, waveletLevel=waveletLevel) for I in range(4) ]
                    else:
                        if basisSet.ndim == 3:
                            self.map = [ OperatorWaveletToFourier(imShape=imShape, samplingSet=samplingSet[:, :, I], basisSet=basisSet[:, :, I], isTransposed=isTransposed, waveletName=waveletName, waveletLevel=waveletLevel) for I in range(4) ]
                        else:
                            self.map = [ OperatorWaveletToFourier(imShape=imShape, samplingSet=samplingSet[:, :, I], basisSet=basisSet, isTransposed=isTransposed, waveletName=waveletName, waveletLevel=waveletLevel) for I in range(4) ]
                else:
                    if basisSet is None:
                        self.map = [ OperatorWaveletToFourier(imShape=imShape, samplingSet=samplingSet, basisSet=None, isTransposed=isTransposed, waveletName=waveletName, waveletLevel=waveletLevel) for I in range(4) ]
                    else:
                        if basisSet.ndim == 3:
                            self.map = [ OperatorWaveletToFourier(imShape=imShape, samplingSet=samplingSet, basisSet=basisSet[:, :, I], isTransposed=isTransposed, waveletName=waveletName, waveletLevel=waveletLevel) for I in range(4) ]
                        else:
                            self.map = [ OperatorWaveletToFourier(imShape=imShape, samplingSet=samplingSet, basisSet=basisSet, isTransposed=isTransposed, waveletName=waveletName, waveletLevel=waveletLevel) for I in range(4) ]
        # Find input shape and output shapes
        inShape = [ self.map[I].inShape for I in range(4) ]
        self.inSlices = None
        if len(inShape[0]) == 2:
            self.inShape = (inShape[0][0], inShape[0][1], 4)
        else:
            self.inShape = np.prod(inShape[0])
            self.inSlices = [ [0, 0] for I in range(4) ]
            self.inSlices[0][1] = inShape[0][0]
            for I in range(1, 4):
                self.inSlices[I][0] = self.inSlices[I-1][1]
                self.inSlices[I][1] = self.inSlices[I][0] + inShape[I][0]
                self.inShape = self.inShape + np.prod(inShape[I])
            self.inShape = (self.inShape,)
        self._inShape = self.inShape
        
        outShape = [ self.map[I].outShape for I in range(4) ]
        self.outSlices = None
        if len(outShape[0])== 2:
            self.outShape = (outShape[0][0], outShape[0][1], 4)
        else:
            self.outShape = np.prod(outShape[0])
            self.outSlices = [ [0, 0] for I in range(4) ]
            self.outSlices[0][1] = outShape[0][0]
            for I in range(1, 4):
                self.outSlices[I][0] = self.outSlices[I-1][1]
                self.outSlices[I][1] = self.outSlices[I][0] + outShape[I][0]
                self.outShape = self.outShape + np.prod(outShape[I])
            self.outShape = (self.outShape,)
        self._outShape = self.outShape

    def eval(self, x, mode=1):
        if mode == 1:
            _y = np.zeros(self._outShape, dtype=np.complex)
            if self.inSlices is None:
                if self.outSlices is None:
                    for I in range(4):
                        _y[:, :, I] = self.map[I].eval(x[:, :, I], mode)
                else:
                    for I in range(4):
                        _y[self.outSlices[I][0]:self.outSlices[I][1]] = self.map[I].eval(x[:, :, I], mode)
            else:
                if self.outSlices is None:
                    for I in range(4):
                        _y[:, :, I] = self.map[I].eval(x[self.inSlices[I][0]:self.inSlices[I][1]], mode)
                else:
                    for I in range(4):
                        _y[self.outSlices[I][0]:self.outSlices[I][1]] = self.map[I].eval(x[self.inSlices[I][0]:self.inSlices[I][1]], mode)
            return _y
        if mode == 2:
            _x = np.zeros(self._inShape, dtype=np.complex)
            if self.outSlices is None:
                if self.inSlices is None:
                    for I in range(4):
                        _x[:, :, I] = self.map[I].eval(x[:, :, I], mode)
                else:
                    for I in range(4):
                        _x[self.inSlices[I][0]:self.inSlices[I][1]] = self.map[I].eval(x[:, :, I], mode)
            else:
                if self.inSlices is None:
                    for I in range(4):
                        _x[:, :, I] = self.map[I].eval(x[self.outSlices[I][0]:self.outSlices[I][1]], mode)
                else:
                    for I in range(4):
                        _x[self.inSlices[I][0]:self.inSlices[I][1]] = self.map[I].eval(x[self.outSlices[I][0]:self.outSlices[I][1]], mode)
            return _x

    def adjoint(self, x):
        return self.eval(x, mode=2)

    def norm(self, maxItns=1E3, absTol=1E-6, relTol=1E-9):
        if self._norm is not None:
            return self._norm        
        # Initialize variables
        x = np.random.normal(size=self._inShape) + 1j * np.random.normal(size=self._inShape)
        x = x / la.norm(x)
        s = 0
        ds = np.inf
        itn = 0
        stop = False
        # Power iteration loop
        while not stop:
            # Evaluate xp = A'A x
            xp = self.adjoint(self.eval(x))
            # Evaluate x' A'Ax to estimate sigma_max^2
            sp = np.sqrt(np.real(np.sum(np.conj(x.flatten()) * xp.flatten())))
            ds = np.abs(sp - s)
            if ds < absTol or ds < absTol * s or itn > maxItns:
                stop = True
            # Normalize singular vector
            x  = xp / la.norm(xp)
            s  = sp
        return s

    def getImageFromWavelet(self, x):
        _im = np.zeros(self.imShape + (4,), dtype=np.complex)
        if self.inSlices is None:
            for I in range(4):
                _im[:, :, I] = self.map[I].getImageFromWavelet(x[:, :, I])
        else:
            for I in range(4):
                _im[:, :, I] = self.map[I].getImageFromWavelet(x[self.inSlices[I][0]:self.inSlices[I][1]])
        return _im

    def getImageFromFourier(self, y):
        _im = np.zeros(self.imShape + (4,), dtype=np.complex)
        if self.outSlices is None:
            for I in range(4):
                _im[:, :, I] = self.map[I].getImageFromFourier(y[:, :, I])
        else:
            for I in range(4):
                _im[:, :, I] = self.map[I].getImageFromFourier(y[self.outSlices[I][0]:self.outSlices[I][1]])
        return _im

    @property
    def shape(self):
        # This is the shape of the matrix representing
        # the operator with vectorized inputs and outputs
        return (np.prod(self._outShape), np.prod(self._inShape))

    def __matmul__(self, x):
        _y = self.eval(np.reshape(x, newshape=self._inShape))
        return _y.ravel()
        
    @property
    def T(self):
        # Instantiate the transpose of the operator
        return OperatorWaveletToFourierX4(imShape=self.imShape, samplingSet=self.samplingSet, basisSet=self.basisSet, isTransposed=not(self.isTransposed), waveletName=self.waveletName, waveletLevel=self.waveletLevel)

    # Create An operator with a 
    def colRestrict(self, basisSet=None):
        # Instantiate operator restricted to some entries
        return OperatorWaveletToFourierX4(imShape=self.imShape, samplingSet=self.samplingSet, basisSet=basisSet, isTransposed=self.isTransposed, waveletName=self.waveletName, waveletLevel=self.waveletLevel)


# OperatorFourierLowRank
#   This implements the linear map that takes as inputs
#   the wavelet coefficients of 4 complex images, and 
#   outputs their Fourier coefficients. The map supports restricting 
#   the support of the wavelet coefficients, and subsampling
#   the Fourier coefficients
class OperatorFourierLowRank(OperatorWaveletToFourierX4):
    def __init__(self, imShape, samplingSet=None, isTransposed=False):
        super().__init__(imShape, samplingSet=samplingSet, basisSet=None, isTransposed=isTransposed, waveletName=None, waveletLevel=None)
        self.mtxShape = (np.prod(imShape), 4)
        self.arrShape = (imShape[0], imShape[1], 4)
        if self.isTransposed:
            self.outShape = (np.prod(imShape), 4)
        else:
            self.inShape = (np.prod(imShape), 4)

    def eval(self, x, mode=1):
        if self.isTransposed:
            if mode == 2:
                return super().eval(np.reshape(x, newshape=self.arrShape), mode)
            else:
                return np.reshape(super().eval(x, mode), newshape=self.mtxShape)
                
        else:
            if mode == 1:
                return super().eval(np.reshape(x, newshape=self.arrShape), mode)
            else:
                return np.reshape(super().eval(x, mode), newshape=self.mtxShape)

    def __matmul__(self, x):
        _y = self.eval(np.reshape(x, newshape=self.inShape))
        return _y.ravel()

    @property
    def T(self):
        # Instantiate the transpose of the operator
        return OperatorFourierLowRank(imShape=self.imShape, samplingSet=self.samplingSet, isTransposed=not(self.isTransposed))

    # Create An operator with a 
    def colRestrict(self, basisSet=None):
        # Instantiate operator restricted to some entries
        raise NotImplementedError('colRestrict is not implemented for OperatorFourierLowRank.')

# %%
def testLinearMap(A, ns=10):
    err_mean = 0.0
    err_std = 0.0
    err_min = np.inf
    err_max = -np.inf
    err_mtx_eval = -np.inf
    err_mtx_adj = -np.inf
    err_adj_eval = -np.inf
    err_adj_adj = -np.inf
    At = A.T
    for I in range(ns):
        x = np.random.normal(size=A.inShape) + 1j * np.random.normal(size=A.inShape)
        y = np.random.normal(size=A.outShape) + 1j * np.random.normal(size=A.outShape)
        Ax = A.eval(x)
        _Ax = At.adjoint(x)
        tAy = A.adjoint(y)
        _tAy = At.eval(y)
        yAx = np.sum( np.conj(y.ravel()) * Ax.ravel() )
        tAyx = np.sum( np.conj(tAy.ravel()) * x.ravel() )
        err = np.abs(yAx - tAyx)
        err_mean = err
        err_std = err ** 2
        err_max = np.maximum(err_max, err)
        err_min = np.minimum(err_min, err)
        err_mtx_eval = np.maximum(err_mtx_eval, la.norm(A @ x.ravel() - Ax.ravel()))
        err_mtx_adj = np.maximum(err_mtx_adj, la.norm(A.T @ y - tAy.ravel()))
        err_adj_eval = np.maximum(err_mtx_eval, la.norm(Ax.ravel() - _Ax.ravel()))
        err_adj_adj = np.maximum(err_mtx_adj, la.norm(_tAy.ravel() - tAy.ravel()))
    err_mean = err_mean / ns
    err_std = np.sqrt(err_std / ns - err_mean ** 2)
    return err_mean, err_std, err_min, err_max, err_mtx_eval, err_mtx_adj, err_adj_eval, err_adj_adj

if __name__ == '__main__':
    ns = 10
    do_sampling = False
    imShape = [ (71, 77), (128, 128) ]
    waveletName = [ 'None', 'haar', 'db4', 'sym5', 'coif5', 'dmey', 'bior2.6', 'rbio2.8', 'dmey' ]
    for _imShape in imShape:
        for _waveletName in waveletName:
            print('-------------------------------------------------------')
            print('Testing operator for image size {:d} x {:d} and wavelet {:s}'.format(_imShape[0], _imShape[1], _waveletName))
            print('-------------------------------------------------------')
            print('\n ****** OperatorWaveletToFourier ***********************')
            if do_sampling:
                delta = np.random.uniform()
                rho = np.random.uniform()
                print('Sampling set ratio (fraction kept):  {:1.3f}'. format(delta))
                print('Basis set ratio (fraction kept):     {:1.3f}'. format(rho))
                samplingSet = np.where(np.random.uniform(size=_imShape) < delta, True, False)
                basisSet = np.where(np.random.uniform(size=getWaveletTransformShape(_imShape, _waveletName)) < rho, True, False)
                A = OperatorWaveletToFourier(_imShape, samplingSet=samplingSet, basisSet=basisSet, isTransposed=False, waveletName=_waveletName)
                inShape = A.inShape
                outShape = A.outShape

                print('     Input shape:   {:d} x 1'.format(inShape[0]))
                print('     Output shape:  {:d} x 1'.format(outShape[0]))
            else:
                A = OperatorWaveletToFourier(_imShape, samplingSet=None, basisSet=None, isTransposed=False, waveletName=_waveletName)
                inShape = A.inShape
                outShape = A.outShape

                print('     Input shape:   {:d} x {:d}'.format(inShape[0], inShape[1]))
                print('     Output shape:  {:d} x {:d}'.format(outShape[0], outShape[1]))

            print('     Operator norm: {:1.5E}'.format(A.norm()))
            print('Testing inner products for {:d} samples...'.format(ns))
            err_mean, err_std, err_min, err_max, err_mtx_eval, err_mtx_adj, err_adj_eval, err_adj_adj = testLinearMap(A)
            print('     Max. Error : {:1.5E}'.format(err_max))
            print('     Min. Error : {:1.5E}'.format(err_min))
            print('     Mean Error : {:1.5E}'.format(err_mean))
            print('          Std   : {:1.5E}'.format(err_std))
            print('     Implementation of __matmul__')
            print('         Max. Error (eval) : {:1.5E}'.format(err_mtx_eval))
            print('         Min. Error (adj)  : {:1.5E}'.format(err_mtx_adj))
            print('     Implementation of .T')
            print('         Max. Error (eval) : {:1.5E}'.format(err_adj_eval))
            print('         Min. Error (adj)  : {:1.5E}'.format(err_adj_adj))

            print('\n ****** OperatorWaveletToFourierX4 *********************')
            if do_sampling:
                delta = np.random.uniform()
                rho = np.random.uniform()
                print('Sampling set ratio (fraction kept):  {:1.3f}'. format(delta))
                print('Basis set ratio (fraction kept):     {:1.3f}'. format(rho))
                samplingSet = np.where(np.random.uniform(size=_imShape) < delta, True, False)
                basisSet = np.where(np.random.uniform(size=getWaveletTransformShape(_imShape, _waveletName)) < rho, True, False)
                A = OperatorWaveletToFourierX4(_imShape, samplingSet=samplingSet, basisSet=basisSet, isTransposed=False, waveletName=_waveletName)
                inShape = A.inShape
                outShape = A.outShape

                print('     Input shape:   {:d} x 1'.format(inShape[0]))
                print('     Output shape:  {:d} x 1'.format(outShape[0]))
            else:
                A = OperatorWaveletToFourierX4(_imShape, samplingSet=None, basisSet=None, isTransposed=False, waveletName=_waveletName)
                inShape = A.inShape
                outShape = A.outShape

                print('     Input shape:   {:d} x {:d}'.format(inShape[0], inShape[1]))
                print('     Output shape:  {:d} x {:d}'.format(outShape[0], outShape[1]))

            print('     Operator norm: {:1.5E}'.format(A.norm()))
            print('Testing inner products for {:d} samples...'.format(ns))
            err_mean, err_std, err_min, err_max, err_mtx_eval, err_mtx_adj, err_adj_eval, err_adj_adj = testLinearMap(A)
            print('     Max. Error : {:1.5E}'.format(err_max))
            print('     Min. Error : {:1.5E}'.format(err_min))
            print('     Mean Error : {:1.5E}'.format(err_mean))
            print('          Std   : {:1.5E}'.format(err_std))
            print('     Implementation of __matmul__')
            print('         Max. Error (eval) : {:1.5E}'.format(err_mtx_eval))
            print('         Min. Error (adj)  : {:1.5E}'.format(err_mtx_adj))
            print('     Implementation of .T')
            print('         Max. Error (eval) : {:1.5E}'.format(err_adj_eval))
            print('         Min. Error (adj)  : {:1.5E}'.format(err_adj_adj))

            print('\n ****** OperatorFourierLowRank *************************')
            if do_sampling:
                delta = np.random.uniform()
                rho = np.random.uniform()
                print('Sampling set ratio (fraction kept):  {:1.3f}'. format(delta))
                # print('Basis set ratio (fraction kept):     {:1.3f}'. format(rho))
                samplingSet = np.where(np.random.uniform(size=_imShape) < delta, True, False)
                # basisSet = np.where(np.random.uniform(size=getWaveletTransformShape(_imShape, _waveletName)) < rho, True, False)
                A = OperatorFourierLowRank(_imShape, samplingSet=samplingSet, isTransposed=False)
                inShape = A.inShape
                outShape = A.outShape

                print('     Input shape:   {:d} x 1'.format(inShape[0]))
                print('     Output shape:  {:d} x 1'.format(outShape[0]))
            else:
                A = OperatorFourierLowRank(_imShape, samplingSet=None, isTransposed=False)
                inShape = A.inShape
                outShape = A.outShape

                print('     Input shape:   {:d} x {:d}'.format(inShape[0], inShape[1]))
                print('     Output shape:  {:d} x {:d}'.format(outShape[0], outShape[1]))

            print('     Operator norm: {:1.5E}'.format(A.norm()))
            print('Testing inner products for {:d} samples...'.format(ns))
            err_mean, err_std, err_min, err_max, err_mtx_eval, err_mtx_adj, err_adj_eval, err_adj_adj = testLinearMap(A)
            print('     Max. Error : {:1.5E}'.format(err_max))
            print('     Min. Error : {:1.5E}'.format(err_min))
            print('     Mean Error : {:1.5E}'.format(err_mean))
            print('          Std   : {:1.5E}'.format(err_std))
            print('     Implementation of __matmul__')
            print('         Max. Error (eval) : {:1.5E}'.format(err_mtx_eval))
            print('         Min. Error (adj)  : {:1.5E}'.format(err_mtx_adj))
            print('     Implementation of .T')
            print('         Max. Error (eval) : {:1.5E}'.format(err_adj_eval))
            print('         Min. Error (adj)  : {:1.5E}'.format(err_adj_adj))            

# %%



import sys,os,math
import numpy as np
import matplotlib.pyplot as plt
sys.path.append('../')
from CSRecoverySuite import get_umask_string, get_method_string, get_wavelet_string
import argparse

home = os.getenv('HOME')

fs = 12
plt.rc('font',  family='serif')
plt.rc('xtick', labelsize='x-small')
plt.rc('ytick', labelsize='x-small')
plt.rc('text',  usetex=True)
start = 0
end   = 15 # last distance to include in plot (final x axis value)
interval = int((end-start)/4) #int(math.ceil(((end-start)/4) / 10.0)) * 10))

def getCorrelationFileName(numsamples, numpts, noise, p, masktype, wavelet, method, useFluidMask):
  res = 'corrcoeff' + str(numpts) + \
        '_noise' + str(int(noise*100)) + \
        '_p' + str(int(p*100)) + masktype + \
        '_n'+str(numsamples)+ \
        '_w' + str(get_wavelet_string(wavelet)) + \
        '_a' + str(get_method_string(method))
  if(useFluidMask):
    res += '_fluid.npy'
  else:
    res += '.npy'
  return res

def get_coeff(noise, p, masktype, wavelet, method, numsamples, numpts, dir, useFluidMask):
  '''
  Retrieve the file with the correlations
  ''' 
  corrfile = dir + get_method_string(method) + '/' + getCorrelationFileName(numsamples, numpts, noise, p, masktype, wavelet, method, useFluidMask)
  if(os.path.isfile(corrfile)):
    coeff = np.load(corrfile)
    return coeff
  else:
    print('Warning: no correlation file found: ',corrfile)
    return None

def plot_corr(noise_percent, p, samptype, wavelet, method, n, num_pts, v, dir, labelstr, useFluidMask):
  coeff = get_coeff(noise_percent, p, samptype, wavelet, method, n, num_pts, dir, useFluidMask)
  if(coeff is None):
    return None
  else:
    coeff = coeff[v]
    coeff[np.isnan(coeff)] = 0
    corravg = np.mean(coeff, axis=1)
    corrmin = np.percentile(coeff, 10, axis=1)
    corrmax = np.percentile(coeff, 90, axis=1)
    size = len(corravg)
    p, = plt.plot(range(start+1,end+1), corravg[start:end], label=labelstr)
    plt.fill_between(range(start+1,end+1), corrmin[start:end], corrmax[start:end], alpha=0.2) # label='10-90 CI')
    return p

def plot_noisediff(args, save_fig=True):
  
  print("Plotting correlations for various k-space noise values...")

  # Set the baseline conditions
  # These are the first element of the lists
  bl_noise   = args.noise[0]
  bl_uval    = args.uval[0]
  bl_utype   = args.utype[0]
  bl_method  = args.method[0]
  bl_wavelet = args.wavelet[0]

  if(args.singlechannel):
    numChannels = 1
  else:
    numChannels = 4
  
  # Loop on the reconstruction component
  for k in range(0,numChannels):

    plt.figure(figsize=(2.2,4))
    # Loop on all noise values
    for noise in args.noise:
      label = r"{}\% noise".format(int(noise*100))
      plot_corr(noise, bl_uval, bl_utype, bl_wavelet, bl_method, args.numsamples, args.numpts, k, args.dir, label, args.usefluidmask)
    
    plt.xlabel('Distance [px]',fontsize=fs)
    plt.ylabel('Correlation Coefficient',fontsize=fs)
    # plt.axhline(y=0,c='gray',lw=1,ls='--',alpha=0.8)
    plt.grid(color='gray', linestyle='-', linewidth=0.5, alpha=0.2)
    plt.tick_params(labelsize=fs)
    plt.ylim([-0.2,1.0])
    plt.xlim([1.0,9.0])
    plt.xticks(np.arange(1, 11, 2))
    plt.legend(loc='upper center',fontsize=fs-2)
    plt.tight_layout()

    if save_fig:
      plt.savefig(args.outputdir + 'diffnoise' + str(start) + 'to' + str(end) + '_p' + str(int(bl_uval*100)) + bl_utype + '_k' + str(k) + '.pdf')
      plt.close()
    else:
      plt.show()

def plot_pdiff(args, save_fig=True):
  
  print("Plotting correlations for various undersampling ratios...")

  # Set the baseline conditions
  # These are the first element of the lists
  bl_noise   = args.noise[0]
  bl_uval    = args.uval[0]
  bl_utype   = args.utype[0]
  bl_method  = args.method[0]
  bl_wavelet = args.wavelet[0]

  if(args.singlechannel):
    numChannels = 1
  else:
    numChannels = 4

  # Loop on the reconstruction component
  for k in range(0,numChannels):

    plt.figure(figsize=(2.2,4))
    # Loop on all noise values
    for p in args.uval:
      label = r"{}\%".format(int(p*100))
      plot_corr(bl_noise, p, bl_utype, bl_wavelet, bl_method, args.numsamples, args.numpts, k, args.dir, label, args.usefluidmask)
    
    plt.xlabel('Distance [px]',fontsize=fs)
    plt.ylabel('Correlation Coefficient',fontsize=fs)
    # plt.axhline(y=0,c='gray',lw=1,ls='--',alpha=0.8)
    plt.grid(color='gray', linestyle='-', linewidth=0.5, alpha=0.2)
    plt.tick_params(labelsize=fs)
    plt.ylim([-0.2,1.0])
    plt.xlim([1.0,9.0])
    plt.xticks(np.arange(1, 11, 2))
    plt.legend(loc='upper center',fontsize=fs-2)
    plt.tight_layout()

    if save_fig:
      plt.savefig(args.outputdir + 'diffuratio' + str(start) + 'to' + str(end) + '_p' + str(int(bl_uval*100)) + bl_utype + '_k' + str(k) + '.pdf')
      plt.close()
    else:
      plt.show()

def plot_sampdiff(args, save_fig=True):
    
  print("Plotting correlations for various undersampling patterns...")

  # Set the baseline conditions
  # These are the first element of the lists
  bl_noise   = args.noise[0]
  bl_uval    = args.uval[0]
  bl_utype   = args.utype[0]
  bl_method  = args.method[0]
  bl_wavelet = args.wavelet[0]

  if(args.singlechannel):
    numChannels = 1
  else:
    numChannels = 4

  # Loop on the reconstruction component
  for k in range(0,numChannels):

    plt.figure(figsize=(2.2,4))
    # Loop on all noise values
    for samptype in args.utype:
      label = get_umask_string(samptype)
      plot_corr(bl_noise, bl_uval, samptype, bl_wavelet, bl_method, args.numsamples, args.numpts, k, args.dir, label, args.usefluidmask)
    
    plt.xlabel('Distance [px]',fontsize=fs)
    plt.ylabel('Correlation Coefficient',fontsize=fs)
    # plt.axhline(y=0,c='gray',lw=1,ls='--',alpha=0.8)
    plt.grid(color='gray', linestyle='-', linewidth=0.5, alpha=0.2)
    plt.tick_params(labelsize=fs)
    plt.ylim([-0.2,1.0])
    plt.xlim([1.0,9.0])
    plt.xticks(np.arange(1, 11, 2))
    plt.legend(loc='upper center',fontsize=fs-2)
    plt.tight_layout()

    if save_fig:
      plt.savefig(args.outputdir + 'diffumask' + str(start) + 'to' + str(end) + '_p' + str(int(bl_uval*100)) + bl_utype + '_k' + str(k) + '.pdf')
      plt.close()
    else:
      plt.show()            

def plot_methoddiff(args, save_fig=True):

  print("Plotting correlations for various reconstruction algorithms...")

  # Set the baseline conditions
  # These are the first element of the lists
  bl_noise   = args.noise[0]
  bl_uval    = args.uval[0]
  bl_utype   = args.utype[0]
  bl_method  = args.method[0]
  bl_wavelet = args.wavelet[0]

  if(args.singlechannel):
    numChannels = 1
  else:
    numChannels = 4

  # Loop on the reconstruction component
  for k in range(0,numChannels):

    plt.figure(figsize=(2.2,4))
    # Loop on all noise values
    for method in args.method:
      label = get_method_string(method)
      plot_corr(bl_noise, bl_uval, bl_utype, bl_wavelet, method, args.numsamples, args.numpts, k, args.dir, label, args.usefluidmask)
    
    plt.xlabel('Distance [px]',fontsize=fs)
    plt.ylabel('Correlation Coefficient',fontsize=fs)
    # plt.axhline(y=0,c='gray',lw=1,ls='--',alpha=0.8)
    plt.grid(color='gray', linestyle='-', linewidth=0.5, alpha=0.2)
    plt.tick_params(labelsize=fs)
    plt.ylim([-0.2,1.0])
    plt.xlim([1.0,9.0])
    plt.xticks(np.arange(1, 11, 2))
    plt.legend(loc='upper center',fontsize=fs-2)
    plt.tight_layout()

    if save_fig:
      plt.savefig(args.outputdir + 'diffmethod' + str(start) + 'to' + str(end) + '_p' + str(int(bl_uval*100)) + bl_utype + '_k' + str(k) + '.pdf')
      plt.close()
    else:
      plt.show()   

def plot_wavediff(args, save_fig=True):

  print("Plotting correlations for various wavelet basis...")

  # Set the baseline conditions
  # These are the first element of the lists
  bl_noise   = args.noise[0]
  bl_uval    = args.uval[0]
  bl_utype   = args.utype[0]
  bl_method  = args.method[0]
  bl_wavelet = args.wavelet[0]

  if(args.singlechannel):
    numChannels = 1
  else:
    numChannels = 4

  # Loop on the reconstruction component
  for k in range(0,numChannels):

    plt.figure(figsize=(2.2,4))
    # Loop on all noise values
    for wavelet in args.wavelet:
      label = get_wavelet_string(wavelet)
      plot_corr(bl_noise, bl_uval, bl_utype, wavelet, bl_method, args.numsamples, args.numpts, k, args.dir, label, args.usefluidmask)
    
    plt.xlabel('Distance [px]',fontsize=fs)
    plt.ylabel('Correlation Coefficient',fontsize=fs)
    plt.grid(color='gray', linestyle='-', linewidth=0.5, alpha=0.2)
    plt.tick_params(labelsize=fs)
    plt.ylim([-0.2,1.0])
    plt.xlim([1.0,9.0])
    plt.xticks(np.arange(1, 11, 2))
    plt.legend(loc='upper center',fontsize=fs-2)
    plt.tight_layout()

    if save_fig:
      plt.savefig(args.outputdir + 'diffwave' + str(start) + 'to' + str(end) + '_p' + str(int(bl_uval*100)) + bl_utype + '_k' + str(k) + '.pdf')
      plt.close()
    else:
      plt.show()   

# MAIN 
if __name__ == '__main__':

  # Init parser
  parser = argparse.ArgumentParser(description='Generate result images.')

  # Load Base Line Params
  # noise
  parser.add_argument('-n', '--noise',
                      action=None,
                      nargs='*',
                      const=None,
                      default=[0.1],
                      type=float,
                      choices=None,
                      required=False,
                      help='list of noise values',
                      metavar='',
                      dest='noise')
  # uval
  parser.add_argument('-u', '--uval',
                      action=None,
                      nargs='*',
                      const=None,
                      default=[0.75],
                      type=float,
                      choices=None,
                      required=False,
                      help='list of undersampling ratios',
                      metavar='',
                      dest='uval')
  # utype
  parser.add_argument('-t', '--utype',
                      action=None,
                      nargs='*',
                      const=None,
                      default=['vardengauss'],
                      type=str,
                      choices=None,
                      required=False,
                      help='list of random undersampling patterns',
                      metavar='',
                      dest='utype')
  # numsamples
  parser.add_argument('-s', '--numsamples',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=100,
                      type=int,
                      choices=None,
                      required=False,
                      help='number of repetitions',
                      metavar='',
                      dest='numsamples')
  # numpts
  parser.add_argument('-np', '--numpts',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=100,
                      type=int,
                      choices=None,
                      required=False,
                      help='number of points for computing the',
                      metavar='',
                      dest='numpts')

  # method
  parser.add_argument('-w', '--wavelet',
                      action=None,
                      nargs='*',
                      const=None,
                      default=['haar'],
                      type=str,
                      choices=['haar','db8'],
                      required=False,
                      help='list of wavelet basis',
                      metavar='',
                      dest='wavelet')

  # method
  parser.add_argument('-m', '--method',
                      action=None,
                      nargs='*',
                      const=None,
                      default=['cs'],
                      type=str,
                      choices=None,
                      required=False,
                      help='list of reconstruction methods',
                      metavar='',
                      dest='method')
  # maindir
  parser.add_argument('-d', '--dir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='folder containing the file with the point locations for the correlation and with subfoldes with the reconstruction methods',
                      metavar='',
                      dest='dir')
  # outputdir
  parser.add_argument('-o', '--outputdir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='output folder for images',
                      metavar='',
                      dest='outputdir')
  # singlechannel
  parser.add_argument('--singlechannel',
                      action='store_true',
                      default=False,
                      required=False,
                      help='treat the image as single-channel, without velocity components',
                      dest='singlechannel') 

  # usefluidmask
  parser.add_argument('--usefluidmask',
                      action='store_true',
                      default=False,
                      required=False,
                      help='compute correlation only within fluid region',
                      dest='usefluidmask')  

  # Print Level
  parser.add_argument('-p', '--printlevel',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=0,
                      type=int,
                      choices=None,
                      required=False,
                      help='print level, 0 - no print, >0 increasingly more information ',
                      metavar='',
                      dest='printlevel')

  # Parse Commandline Arguments
  args = parser.parse_args()

  # Plot the parameter perturbations
  if(len(args.noise) > 1):
    plot_noisediff(args)
  if(len(args.uval) > 1):
    plot_pdiff(args)
  if(len(args.utype) > 1):
    plot_sampdiff(args)
  if(len(args.method) > 1):
    plot_methoddiff(args)
  if(len(args.method) > 1):
    plot_wavediff(args)

  # Completed!
  if(args.printlevel > 0):
    print('Completed!!!')  



import sys
import os
import warnings
sys.path.append('../')
from recover import linear_reconstruction, recover_vel
from genSamples import getKspace
import numpy as np
import numpy.fft as fft
import matplotlib.pyplot as plt
from matplotlib.ticker import PercentFormatter, ScalarFormatter
from CSRecoverySuite import crop, isvalidcrop, extractFluidMask
from CSRecoverySuite import get_umask_string, get_method_string, get_wavelet_string
import argparse
from datetime import datetime

warnings.filterwarnings('error')

home = os.getenv('HOME')
hatchPatterns = ['-', '+', 'x', '\\', '*', 'o', 'O', '.']

fs = 12
plt.rc('font',  family='serif')
plt.rc('xtick', labelsize='x-small')
plt.rc('ytick', labelsize='x-small')
plt.rc('text',  usetex=True)

def generateViolinPlot(lgd,xlabel,allplt,allplt_lin=None):
  fig, ax = plt.subplots(figsize=(3,4))

  bplts = plt.violinplot(allplt,showmeans=False, showmedians=False)
  for patch in bplts['bodies']:
      patch.set_facecolor('r')
      patch.set_edgecolor('r')
      patch.set_linewidth(0.8)
      patch.set_alpha(0.5)

  if(not(allplt_lin is None)):
    bplts = plt.violinplot(allplt_lin,showmeans=False, showmedians=False)
    for patch in bplts['bodies']:
      patch.set_facecolor('gray')
      patch.set_edgecolor('gray')
      patch.set_linewidth(0.8)
      patch.set_alpha(0.5)
  
  plt.ylabel('MSE',fontsize=fs)
  plt.tick_params(labelsize=fs)
  ax.set_xticks(range(1, len(lgd)+1))
  ax.set_xticklabels(lgd)
  plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0), useOffset=True)
  plt.xlabel(xlabel)
  plt.tight_layout()

  return fig

def generateBarPlot(lgd,xlabel,ylabel,
                    usecompleximgs,addlinearrec,
                    allplt_mag,allplt_ang,allplt_cmx,
                    allplt_mag_lin=None,allplt_ang_lin=None,allplt_cmx_lin=None,
                    rotext=False):

  # Set Space and Initial Shift for the plots
  pltShift = 1.0
  pltNum = 4
  if(addlinearrec):
    if(usecompleximgs):
      pltShift = 2.5
      pltNum = 7 
    else:
      pltShift = 0.5
      pltNum = 2
  else:
    if(usecompleximgs):
      pltShift = 1.0
      pltNum = 4
    else:
      pltShift = 0.0
      pltNum = 2

  # New picture
  fig, ax = plt.subplots(figsize=(4,3))

  # Set Values on the X Axis
  xVals_mag = np.arange(0,pltNum*len(allplt_mag),pltNum)
  if(usecompleximgs):
    xVals_ang = np.arange(1,pltNum*len(allplt_ang),pltNum)
    xVals_cmx = np.arange(2,pltNum*len(allplt_cmx),pltNum)
  
  # Compute Mean values
  # Magnitude
  yVals_mag = np.zeros(len(allplt_mag))
  for loopA in range(len(allplt_mag)):
    yVals_mag[loopA] = allplt_mag[loopA].mean()
  
  # Angle
  yVals_ang = np.zeros(len(allplt_ang))
  for loopA in range(len(allplt_ang)):
    yVals_ang[loopA] = allplt_ang[loopA].mean()

  # Complex
  yVals_cmx = np.zeros(len(allplt_cmx))
  for loopA in range(len(allplt_cmx)):
    yVals_cmx[loopA] = allplt_cmx[loopA].mean()
  
  # Compute Errors
  # Magnitude
  yErr_mag = np.zeros((2,len(allplt_mag)))
  for loopA in range(len(allplt_mag)):
    yErr_mag[0,loopA] = np.abs(np.percentile(allplt_mag[loopA],10)-yVals_mag[loopA])
    yErr_mag[1,loopA] = np.abs(np.percentile(allplt_mag[loopA],90)-yVals_mag[loopA])
  # Angle
  yErr_ang = np.zeros((2,len(allplt_ang)))
  for loopA in range(len(allplt_ang)):
    yErr_ang[0,loopA] = np.abs(np.percentile(allplt_ang[loopA],10)-yVals_ang[loopA])
    yErr_ang[1,loopA] = np.abs(np.percentile(allplt_ang[loopA],90)-yVals_ang[loopA])
  # Complex
  yErr_cmx = np.zeros((2,len(allplt_cmx)))
  for loopA in range(len(allplt_cmx)):
    yErr_cmx[0,loopA] = np.abs(np.percentile(allplt_cmx[loopA],10)-yVals_cmx[loopA])
    yErr_cmx[1,loopA] = np.abs(np.percentile(allplt_cmx[loopA],90)-yVals_cmx[loopA])

  # Plot MSE
  ax.bar(xVals_mag, yVals_mag, yerr=yErr_mag, width=0.8, error_kw=dict(lw=0.5, capsize=2, capthick=0.5),label='cs mag',hatch='++')
  if(usecompleximgs):
    ax.bar(xVals_ang, yVals_ang, yerr=yErr_ang, width=0.8, error_kw=dict(lw=0.5, capsize=2, capthick=0.5),label='cs ang',hatch='--')
    ax.bar(xVals_cmx, yVals_cmx, yerr=yErr_cmx, width=0.8, error_kw=dict(lw=0.5, capsize=2, capthick=0.5),label='cs cmx',hatch='\\\\')

  # Plot MSE for linear reconstruction
  if(addlinearrec):
    
    # Set X values
    if(usecompleximgs):
      xVals_mag = np.arange(3,pltNum*len(allplt_mag_lin),pltNum)
      xVals_ang = np.arange(4,pltNum*len(allplt_ang_lin),pltNum)
      xVals_cmx = np.arange(5,pltNum*len(allplt_cmx_lin),pltNum)
    else:
      xVals_mag = np.arange(1,pltNum*len(allplt_mag_lin),pltNum)

    # Compute Mean values
    yVals_mag = np.zeros(len(allplt_mag_lin))
    for loopA in range(len(allplt_mag_lin)):
      yVals_mag[loopA] = allplt_mag_lin[loopA].mean()

    #
    yVals_ang = np.zeros(len(allplt_ang_lin))
    for loopA in range(len(allplt_ang_lin)):
      yVals_ang[loopA] = allplt_ang_lin[loopA].mean()

    #
    yVals_cmx = np.zeros(len(allplt_cmx_lin))
    for loopA in range(len(allplt_cmx_lin)):
      yVals_cmx[loopA] = allplt_cmx_lin[loopA].mean()

    # Compute Errors
    yErr_mag = np.zeros((2,len(allplt_mag_lin)))
    for loopA in range(len(allplt_mag_lin)):
      yErr_mag[0,loopA] = np.abs(np.percentile(allplt_mag_lin[loopA],10)-yVals_mag[loopA])
      yErr_mag[1,loopA] = np.abs(np.percentile(allplt_mag_lin[loopA],90)-yVals_mag[loopA])
    #
    yErr_ang = np.zeros((2,len(allplt_ang_lin)))
    for loopA in range(len(allplt_ang_lin)):
      yErr_ang[0,loopA] = np.abs(np.percentile(allplt_ang_lin[loopA],10)-yVals_ang[loopA])
      yErr_ang[1,loopA] = np.abs(np.percentile(allplt_ang_lin[loopA],90)-yVals_ang[loopA])
    #
    yErr_cmx = np.zeros((2,len(allplt_cmx_lin)))
    for loopA in range(len(allplt_cmx_lin)):
      yErr_cmx[0,loopA] = np.abs(np.percentile(allplt_cmx_lin[loopA],10)-yVals_cmx[loopA])
      yErr_cmx[1,loopA] = np.abs(np.percentile(allplt_cmx_lin[loopA],90)-yVals_cmx[loopA])

    ax.bar(xVals_mag, yVals_mag, yerr=yErr_mag, width=0.8, error_kw=dict(lw=0.5, capsize=2, capthick=0.5),label='lin mag',hatch='xx')
    if(usecompleximgs):
      ax.bar(xVals_ang, yVals_ang, yerr=yErr_ang, width=0.8, error_kw=dict(lw=0.5, capsize=2, capthick=0.5),label='lin ang',hatch='oo')
      ax.bar(xVals_cmx, yVals_cmx, yerr=yErr_cmx, width=0.8, error_kw=dict(lw=0.5, capsize=2, capthick=0.5),label='lin cmx',hatch='..')
   
  # Set X thicks
  ax.set_xticks(pltShift+pltNum*np.arange(len(lgd)))

  plt.ylabel(ylabel,fontsize=fs)
  if(rotext):
    ax.set_xticklabels(lgd, rotation=40, ha='right')  
  else:
    ax.set_xticklabels(lgd)
  plt.ticklabel_format(style='sci', axis='y', scilimits=(0,0), useOffset=True)
  plt.legend(loc='lower center', fontsize=fs-4, bbox_to_anchor=(0.5, 1.0), 
             fancybox=False, shadow=False, ncol=3, handlelength=1.5, 
             labelspacing=0.2, columnspacing=5.25)
  plt.yscale('log')
  plt.tick_params(axis='both', reset=True, grid_color='gray', grid_alpha=0.1, 
                  which='both',bottom=True,left=True,top=False,right=False)
  plt.grid(True, axis='both', which='both')
  plt.rc('axes', axisbelow=True)
  plt.xlabel(xlabel)
  plt.tight_layout()

  return fig

def get_files(dir, maskdir, noise, uval, utype, wavelet, method, numsamples, usemultipatterns):
  # Get Image File Names
  orig_file        = dir + 'imgs_n1.npy'
  fourier_file     = dir + 'noisy_noise' + str(int(noise*100)) + '_n' + str(numsamples) + '.npy'
  if(usemultipatterns):
    mask_file = maskdir + 'undersamplpattern_p' + str(int(uval*100)) + utype + '_n' + str(numsamples) + '.npy'
  else:
    mask_file = maskdir + 'undersamplpattern_p' + str(int(uval*100)) + utype + '.npy'
  recovered_file   = dir + get_method_string(method) + '/' + 'rec_noise' + str(int(noise*100)) + \
                                          '_p' + str(int(uval*100)) + utype + \
                                          '_n' + str(numsamples) + \
                                          '_w' + str(get_wavelet_string(wavelet)) + \
                                          '_a' + str(get_method_string(method)) + '.npy'

  # print('orig_file',orig_file)
  # print('fourier_file',fourier_file)
  # print('mask_file',mask_file)
  # print('recovered_file',recovered_file)

  # Check if files exist and open the images
  # Original image
  if(os.path.exists(orig_file)):
    try:
      orig = np.load(orig_file).astype(np.complex)
    except Warning:
      print('ERROR: Cannot read original image - ',orig)
      sys.exit(-1)
    if(not(isvalidcrop(orig))):
      print('ERROR: Invalid original image file.')
      sys.exit(-1)
  else:
    print('ERROR: original image file not found: ',orig_file)
    sys.exit(-1)

  # Noisy Fourier Image
  if(os.path.exists(fourier_file)):
    try:
      fourier = np.load(fourier_file).astype(np.complex)
    except Warning:
      print('ERROR: Cannot read Fourier file - ',fourier_file)
      sys.exit(-1)
  else:
    print('ERROR: file for noisy Fourier image not found: ',fourier_file)
    sys.exit(-1)

  # Undersampling Mask
  if(os.path.exists(mask_file)):
    try:
      mask = np.load(mask_file).astype(bool)
    except Warning:
      print('ERROR: Cannot read binary mask - ',mask_file)
      sys.exit(-1)
  else:
    print('ERROR: undersampling mask file not found: ',mask_file)
    sys.exit(-1)

  # File with reconstructed image
  if(os.path.exists(recovered_file)):
    try:
      recovered = np.load(recovered_file).astype(np.complex)
    except Warning:
      print('ERROR: Cannot read reconstructedd image - ',recovered_file)
      sys.exit(-1)
  else:
    print('ERROR: file for reconstructed images not found: ',recovered_file)
    sys.exit(-1)

  return fourier, orig, mask, recovered

def get_complex(channel, dir, maskdir, noise, uval, utype, wavelet, method, numsamples, addlinearrec, usemultipatterns):
    
  # Get all files
  fourier, orig, mask, recovered = get_files(dir, maskdir, noise, uval, utype, wavelet, method, numsamples, usemultipatterns)

  # Get velocity encoding
  venc_file = dir + 'venc_n1.npy'
  if os.path.exists(venc_file):
    venc = np.load(dir + 'venc_n1.npy').astype(np.float)
  else:
    venc = None
    print('WARNING: file for velocity encoding not found: ',venc_file)  

  # Compute linear reconstructions if requested
  retLinRec = None
  if(addlinearrec):
    # Perform linear reconstruction
    linrec = linear_reconstruction(fourier,mask)
    # Return complex images only for the requested channel
    retLinRec = linrec[:,channel,:,:,:]

  # Convert True Image to Complex
  kspace_orig = np.zeros(orig.shape,dtype=complex)
  referencephase = np.zeros(orig[:,0].shape)
  magnitudes = np.ones(orig.shape)
  mag = orig[0,0]
  refphase = referencephase[0]
  vel = orig[0,1:4]
  compyk = np.zeros(vel.shape, dtype=complex)
  for k in range(orig.shape[2]):
      kspace_orig[0,0,k] = mag[k]*np.exp(1j*refphase[k])
      for j in range(0,orig.shape[1]-1):
          kspace_orig[0,j+1,k] = magnitudes[0,j+1,k]*np.exp(1j*refphase[k])*np.exp(np.pi*1j*vel[j,k]/venc)

  # Return complex images only for the requested channel
  return recovered[:,channel,:,:,:], kspace_orig[:,channel,:,:,:], retLinRec

def get_final(channel, dir, maskdir, noise, uval, utype, wavelet, method, numsamples, addlinearrec, usemultipatterns):

  # Get all files
  fourier, orig, mask, recovered = get_files(dir, maskdir, noise, uval, utype, wavelet, method, numsamples, usemultipatterns)

  # Get velocity encoding
  venc_file = dir + 'venc_n1.npy'
  if os.path.exists(venc_file):
    venc = np.load(dir + 'venc_n1.npy').astype(np.float)
  else:
    venc = None
    print('WARNING: file for velocity encoding not found: ',venc_file)

  # Recover the velocity components  
  csvels = recover_vel(recovered, venc)

  # Compute linear reconstructions if requested
  retLinVels = None
  if(addlinearrec):
    # Perform linear reconstruction
    linrec = linear_reconstruction(fourier,mask)
    # Compute velocities from linear reconstruction
    linvels = recover_vel(linrec, venc)
    retLinVels = linvels[:,channel,:,:,:]
    
  # Return
  return csvels[:,channel,:,:,:], orig[:,channel,:,:,:], retLinVels

def get_perc_diff(k,csimgs,refimg,useFluidMask=False,fluidMask=None):
  '''
  Evaluate the percentage 
  '''
  refimg[0,0][np.abs(refimg[0,0]) < 1.0e-3] = 1.0e-3
  res = np.absolute((np.absolute(refimg[0,0])-np.absolute(csimgs[k,0]))/(np.absolute(refimg[0,0])))
  if(useFluidMask):
    res = np.median(res[fluidMask])
  else:
    res = np.median(res)
  return res

def get_mse_mag(k,csimgs,refimg,useFluidMask=False,fluidMask=None):
  if(useFluidMask):
    imgNum = (np.absolute(refimg[0,0])-np.absolute(csimgs[k,0]))**2
    imgDen = np.absolute(refimg[0,0])**2
    res = np.mean(imgNum[fluidMask])
    den = np.mean(imgDen[fluidMask])
    if(den > 1.0e-12):
      res = res/den
  else:
    res = np.mean((np.absolute(refimg[0,0])-np.absolute(csimgs[k,0]))**2)    
    if(np.mean(np.absolute(refimg[0,0])**2) > 1.0e-12):
      res = res/np.mean(np.absolute(refimg[0,0])**2)    

  return res

def get_mse_ang(k,csimgs,refimg,useFluidMask=False,fluidMask=None):
  if(useFluidMask):
    imgNum = (np.angle(refimg[0,0])-np.angle(csimgs[k,0]))**2
    res = np.mean(imgNum[fluidMask])
  else:
    res = np.mean((np.angle(refimg[0,0])-np.angle(csimgs[k,0]))**2)   
    # mse_ang_cs[k]  = mse_ang_cs[k]/np.mean(np.angle(refimg)**2)

  return res

def get_mse_cmx(k,csimgs,refimg,useFluidMask=False,fluidMask=None):
  if(useFluidMask):
    imgNum = (np.absolute(refimg[0,0]-csimgs[k,0]))**2
    imgDen = np.absolute(refimg[0,0])**2
    res = np.mean(imgNum[fluidMask])
    den = np.mean(imgDen[fluidMask])

    if(den > 1.0e-12):
      res = res/den    
  else:

    res = np.mean((np.absolute(refimg[0,0]-csimgs[k,0]))**2)
    if(np.mean(np.absolute(refimg[0,0])**2) > 1.0e-12):
      res = res/np.mean(np.absolute(refimg[0,0])**2)

  return res

def get_error(channel, dir, maskdir, noise, uval, utype, wavelet, method, numsamples, usecompleximgs, usetrueimg, addlinearrec, useFluidMask, fluidMaskFile, usemultipatterns):
    
  # Get CS reconstructions, linear reconstruction and original image
  if usecompleximgs:
    csimgs, orig, linimgs = get_complex(channel, dir, maskdir, noise, uval, utype, wavelet, method, numsamples, addlinearrec, usemultipatterns)
  else:
    csimgs, orig, linimgs = get_final(channel, dir, maskdir, noise, uval, utype, wavelet, method, numsamples, addlinearrec, usemultipatterns) 

  # If use fluid mask, then compute the mask using the original image density+velocities
  fluidMask = None
  if(useFluidMask):

    if(fluidMaskFile == ''):
      orig_forMask = np.load(dir + 'imgs_n1.npy').astype(np.complex)
      fluidMask = extractFluidMask(orig_forMask)
      print('Using automatically extracted fluid mask...')
    else:
      fluidMask = np.load(dir + fluidMaskFile)
      print('Using fluid mask from file: '+ dir + fluidMaskFile)

  # Get average cs and linear reconstructed image
  if(usetrueimg):
    refimg = orig
  else:
    refimg = csimgs.mean(axis=0,keepdims=True)

  # DEBUG - CHECK THE DIFFERENCE IN RECONSTRUCTED IMAGE
  # plt.figure(figsize=(10,3))
  # print('noise ' + str(noise) + ' uval ' + str(uval) + ' utype ' + str(utype) + ' wave ' + str(wavelet) + ' alg ' + str(method))
  # if usecompleximgs:
  #   print('Using Complex Images')
  # else:
  #   print('Using Components')
  # plt.subplot(1,3,1)
  # plt.title('TRUE')
  # plt.imshow(np.absolute(refimg[0,0]))
  # plt.colorbar()
  # plt.subplot(1,3,2)
  # plt.title('REC')
  # plt.imshow(np.absolute(csimgs[0,0]))
  # plt.colorbar()
  # plt.subplot(1,3,3)
  # plt.title('LIN')
  # plt.imshow(np.absolute(linimgs[0,0]))
  # plt.colorbar()  
  # plt.show()
  # plt.close()

  # print('Sizes')
  # print('csimgs:',csimgs.shape)
  # print('orig: ',orig.shape)
  # print('linimgs: ',linimgs.shape)

  # Init vectors for MSE: one for every sample
  mse_mag_cs  = np.zeros(csimgs.shape[0])
  mse_ang_cs  = np.zeros(csimgs.shape[0])
  mse_cmx_cs  = np.zeros(csimgs.shape[0])
  perc_err_cs = np.zeros(csimgs.shape[0])

  if(addlinearrec):
    
    # Init vectors for linear MSE
    mse_mag_lin  = np.zeros(csimgs.shape[0])
    mse_ang_lin  = np.zeros(csimgs.shape[0])
    mse_cmx_lin  = np.zeros(csimgs.shape[0])
    perc_err_lin = np.zeros(csimgs.shape[0])

    # Compute image average
    avglin = linimgs.mean(axis=0,keepdims=True)
    
  for k in range(csimgs.shape[0]):

    # Magnitude
    mse_mag_cs[k]  = get_mse_mag(k,csimgs,refimg,useFluidMask,fluidMask)
    # Angle - Now that you mention it, I think we should add np.angle(orig)-np.angle(csimgs[k]) and np.abs(orig-csimgs[k]).
    mse_ang_cs[k]  = get_mse_ang(k,csimgs,refimg,useFluidMask,fluidMask)
    # Complex
    mse_cmx_cs[k]  = get_mse_cmx(k,csimgs,refimg,useFluidMask,fluidMask)
    # Percent Error
    perc_err_cs[k] = get_perc_diff(k,csimgs,refimg,useFluidMask,fluidMask)

    if(addlinearrec):
      
      # Magnitude
      mse_mag_lin[k] = get_mse_mag(k,linimgs,refimg,useFluidMask,fluidMask)
      # Angle
      mse_ang_lin[k] = get_mse_ang(k,linimgs,refimg,useFluidMask,fluidMask)
      # Complex
      mse_cmx_lin[k] = get_mse_cmx(k,linimgs,refimg,useFluidMask,fluidMask)
      # Percent Error
      perc_err_lin[k] = get_perc_diff(k,csimgs,refimg,useFluidMask,fluidMask)

  if(addlinearrec):
    return mse_mag_cs, mse_ang_cs, mse_cmx_cs, perc_err_cs, mse_mag_lin, mse_ang_lin, mse_cmx_lin, perc_err_lin
  else:
    return mse_mag_cs, mse_ang_cs, mse_cmx_cs, perc_err_cs, None, None, None, None

def plot_pdiff(args,perc_dict):

  print('Plotting MSE for various undersampling ratios...')

  # Set the baseline conditions
  # These are the first element of the lists
  bl_noise  = args.noise[0]
  bl_uval   = args.uval[0]
  bl_utype  = args.utype[0]
  bl_method = args.method[0]
  bl_wavelet = args.wavelet[0]

  if(args.singlechannel):
    numChannels = 1
  else:
    numChannels = 4

  for ch in range(numChannels):

    # CS    
    allplt_mag = [None]*len(args.uval)
    allplt_ang = [None]*len(args.uval)
    allplt_cmx = [None]*len(args.uval)
    # Linear
    allplt_mag_lin = [None]*len(args.uval)
    allplt_ang_lin = [None]*len(args.uval)
    allplt_cmx_lin = [None]*len(args.uval)

    for i,uval in enumerate(sorted(args.uval)): 
      
      mse_mag_cs, mse_ang_cs, mse_cmx_cs, perc_err_cs, mse_mag_lin, mse_ang_lin, mse_cmx_lin, perc_err_lin = get_error(ch, args.dir, args.maskdir, bl_noise, uval, bl_utype, bl_wavelet, bl_method, args.numsamples, args.usecompleximgs, args.usetrueimg, args.addlinearrec, args.usefluidmask, args.fluidmaskfile, args.usemultipatterns)
      
      allplt_mag[i] = mse_mag_cs
      allplt_ang[i] = mse_ang_cs
      allplt_cmx[i] = mse_cmx_cs

      if(args.addlinearrec):
        allplt_mag_lin[i] = mse_mag_lin
        allplt_ang_lin[i] = mse_ang_lin
        allplt_cmx_lin[i] = mse_cmx_lin

      if(args.usetrueimg):
        msg = 'tru'
        ylabel = 'MSE - w.r.t. true image'
      else:
        msg = 'avg'
        ylabel = 'MSE - w.r.t. average image'

      if(args.usecompleximgs):
        msg += '_cmx'
      else:
        msg += '_vel'

      # Store percent error
      comp = 'pdiff_k'+str(ch)+'_p'+str(uval)+'_'+msg
      perc_dict[comp+'_cs'] = perc_err_cs
      perc_dict[comp+'_lin'] = perc_err_lin

    lgd = [str(int(x*100)) + '\%' for x in sorted(args.uval)]
    fig = generateBarPlot(lgd,'Undersampling ratio',ylabel,
                          args.usecompleximgs,args.addlinearrec,
                          allplt_mag,allplt_ang,allplt_cmx,
                          allplt_mag_lin,allplt_ang_lin,allplt_cmx_lin)

    fname = args.outputdir + 'pdiff_' + msg + '_noise' + str(int(bl_noise*100)) + '_' + bl_utype + '_k' + str(ch) + '.png'
    plt.savefig(fname,dpi=200)
    print("MSE Image saved: " + fname)
    plt.close(fig)

def plot_noisediff(args,perc_dict):
  '''
  Ploting MSE for variable k-space noise
  '''

  print('Plotting MSE for various noise intensities...')

  # Set the baseline conditions
  # These are the first element of the lists
  bl_noise  = args.noise[0]
  bl_uval   = args.uval[0]
  bl_utype  = args.utype[0]
  bl_method = args.method[0]
  bl_wavelet = args.wavelet[0]

  if(args.singlechannel):
    numChannels = 1
  else:
    numChannels = 4

  for ch in range(numChannels):

    # CS    
    allplt_mag = [None]*len(args.noise)
    allplt_ang = [None]*len(args.noise)
    allplt_cmx = [None]*len(args.noise)
    # Linear
    allplt_mag_lin = [None]*len(args.noise)
    allplt_ang_lin = [None]*len(args.noise)
    allplt_cmx_lin = [None]*len(args.noise)

    for i,noise in enumerate(sorted(args.noise)):
      
      mse_mag_cs, mse_ang_cs, mse_cmx_cs, perc_err_cs, mse_mag_lin, mse_ang_lin, mse_cmx_lin, perc_err_lin = get_error(ch, args.dir, args.maskdir, noise, bl_uval, bl_utype, bl_wavelet, bl_method, args.numsamples, args.usecompleximgs, args.usetrueimg, args.addlinearrec, args.usefluidmask, args.fluidmaskfile, args.usemultipatterns)

      allplt_mag[i] = mse_mag_cs
      allplt_ang[i] = mse_ang_cs
      allplt_cmx[i] = mse_cmx_cs

      if(args.addlinearrec):
        allplt_mag_lin[i] = mse_mag_lin
        allplt_ang_lin[i] = mse_ang_lin
        allplt_cmx_lin[i] = mse_cmx_lin
        
      if(args.usetrueimg):
        msg = 'tru'
        ylabel = 'MSE - w.r.t. true image'
      else:
        msg = 'avg'
        ylabel = 'MSE - w.r.t. average image'

      if(args.usecompleximgs):
        msg += '_cmx'
      else:
        msg += '_vel'

      # Store percent error
      comp = 'noisediff_k'+str(ch)+'_noise'+str(noise)+'_'+msg
      perc_dict[comp+'_cs'] = perc_err_cs
      perc_dict[comp+'_lin'] = perc_err_lin

    # Generate Plots
    lgd = [str(int(x*100)) + '\%' for x in sorted(args.noise)]
    # fig = generateViolinPlot(lgd,'Noise intensity',allplt,allplt_lin)
    fig = generateBarPlot(lgd,'Noise intensity',ylabel,
                          args.usecompleximgs,args.addlinearrec,
                          allplt_mag,allplt_ang,allplt_cmx,
                          allplt_mag_lin,allplt_ang_lin,allplt_cmx_lin)

    # Save Img  
    fname = args.outputdir + 'noisediff_' + msg + '_p' + str(int(bl_uval*100)) + '_' + bl_utype + '_k' + str(ch) + '.png'
    plt.savefig(fname,dpi=200)
    print("MSE Image saved: " + fname)
    plt.close(fig)

def plot_maskdiff(args,perc_dict):
  '''
  Ploting MSE for variable undesampling masks
  '''
  print('Plotting MSE for various undersampling masks...')

  # Set the baseline conditions
  # These are the first element of the lists
  bl_noise  = args.noise[0]
  bl_uval   = args.uval[0]
  bl_utype  = args.utype[0]
  bl_method = args.method[0]
  bl_wavelet = args.wavelet[0]

  if(args.singlechannel):
    numChannels = 1
  else:
    numChannels = 4

  for ch in range(numChannels):

    # CS    
    allplt_mag = [None]*len(args.utype)
    allplt_ang = [None]*len(args.utype)
    allplt_cmx = [None]*len(args.utype)
    # Linear
    allplt_mag_lin = [None]*len(args.utype)
    allplt_ang_lin = [None]*len(args.utype)
    allplt_cmx_lin = [None]*len(args.utype)

    for i,utype in enumerate(args.utype):
      
      mse_mag_cs, mse_ang_cs, mse_cmx_cs, perc_err_cs, mse_mag_lin, mse_ang_lin, mse_cmx_lin, perc_err_lin = get_error(ch,args.dir, args.maskdir, bl_noise, bl_uval, utype, bl_wavelet, bl_method, args.numsamples, args.usecompleximgs, args.usetrueimg, args.addlinearrec, args.usefluidmask, args.fluidmaskfile, args.usemultipatterns)
      
      allplt_mag[i] = mse_mag_cs
      allplt_ang[i] = mse_ang_cs
      allplt_cmx[i] = mse_cmx_cs

      if(args.addlinearrec):
        allplt_mag_lin[i] = mse_mag_lin
        allplt_ang_lin[i] = mse_ang_lin
        allplt_cmx_lin[i] = mse_cmx_lin

      if(args.usetrueimg):
        msg = 'tru'
        ylabel = 'MSE - w.r.t. true image'
      else:
        msg = 'avg'
        ylabel = 'MSE - w.r.t. average image'

      if(args.usecompleximgs):
        msg += '_cmx'
      else:
        msg += '_vel'

      # Store percent error
      comp = 'maskdiff_k'+str(ch)+'_'+str(utype)+'_'+msg
      perc_dict[comp+'_cs'] = perc_err_cs
      perc_dict[comp+'_lin'] = perc_err_lin

    lgd = []
    for loopA in range(len(args.utype)):
      lgd.append(get_umask_string(args.utype[loopA]))

    # fig = generateViolinPlot(lgd,'Undersampling mask',allplt,allplt_lin)    
    fig = generateBarPlot(lgd,'Undersampling mask',ylabel,
                          args.usecompleximgs,args.addlinearrec,
                          allplt_mag,allplt_ang,allplt_cmx,
                          allplt_mag_lin,allplt_ang_lin,allplt_cmx_lin,rotext=True)

    # Save Img  
    fname = args.outputdir + 'maskdiff_' + msg + '_noise' + str(int(bl_noise*100)) + '_p' + str(int(bl_uval*100)) + '_' + bl_utype + '_k' + str(ch) + '.png'
    plt.savefig(fname,dpi=200)
    print("MSE Image saved: " + fname)
    plt.close(fig)  

def plot_methoddiff(args,perc_dict):

  print('Plotting MSE for various reconstruction methods...')
    
  # Set the baseline conditions
  # These are the first element of the lists
  bl_noise  = args.noise[0]
  bl_uval   = args.uval[0]
  bl_utype  = args.utype[0]
  bl_method = args.method[0]
  bl_wavelet = args.wavelet[0]

  if(args.singlechannel):
    numChannels = 1
  else:
    numChannels = 4

  for ch in range(numChannels):

    # CS    
    allplt_mag = [None]*len(args.method)
    allplt_ang = [None]*len(args.method)
    allplt_cmx = [None]*len(args.method)
    # Linear
    allplt_mag_lin = [None]*len(args.method)
    allplt_ang_lin = [None]*len(args.method)
    allplt_cmx_lin = [None]*len(args.method)

    for i,method in enumerate(args.method):

      mse_mag_cs, mse_ang_cs, mse_cmx_cs, perc_err_cs, mse_mag_lin, mse_ang_lin, mse_cmx_lin, perc_err_lin = get_error(ch,args.dir, args.maskdir, bl_noise, bl_uval, bl_utype, bl_wavelet, method, args.numsamples, args.usecompleximgs, args.usetrueimg, args.addlinearrec, args.usefluidmask, args.fluidmaskfile, args.usemultipatterns)
        
      allplt_mag[i] = mse_mag_cs
      allplt_ang[i] = mse_ang_cs
      allplt_cmx[i] = mse_cmx_cs

      if(args.addlinearrec):
        allplt_mag_lin[i] = mse_mag_lin
        allplt_ang_lin[i] = mse_ang_lin
        allplt_cmx_lin[i] = mse_cmx_lin

      if(args.usetrueimg):
        msg = 'tru'
        ylabel = 'MSE - w.r.t. true image'
      else:
        msg = 'avg'
        ylabel = 'MSE - w.r.t. average image'

      if(args.usecompleximgs):
        msg += '_cmx'
      else:
        msg += '_vel'

      # Store percent error
      comp = 'methoddiff_k'+str(ch)+'_'+str(method)+'_'+msg
      perc_dict[comp+'_cs'] = perc_err_cs
      perc_dict[comp+'_lin'] = perc_err_lin        

    # Generate Plots
    lgd = []
    for loopA in range(len(args.method)):
      lgd.append(get_method_string(args.method[loopA]))

    # fig = generateViolinPlot(lgd,'Reconstruction method',allplt,allplt_lin)    
    fig = generateBarPlot(lgd,'Reconstruction method',ylabel,
                          args.usecompleximgs,args.addlinearrec,
                          allplt_mag,allplt_ang,allplt_cmx,
                          allplt_mag_lin,allplt_ang_lin,allplt_cmx_lin,rotext=True)

    fname = args.outputdir + 'methoddiff_' + msg + '_noise' + str(int(bl_noise*100)) + '_p'+str(int(bl_uval*100)) + '_' + bl_utype + '_k' + str(ch) + '.png'
    plt.savefig(fname,dpi=200)
    print("MSE Image saved: " + fname)
    plt.close(fig)

# Difference in the wavelet frame
def plot_waveletdiff(args,perc_dict):

  print('Plotting MSE for various wavelet frames...')
    
  # Set the baseline conditions
  # These are the first element of the lists
  bl_noise   = args.noise[0]
  bl_uval    = args.uval[0]
  bl_utype   = args.utype[0]
  bl_method  = args.method[0]
  bl_wavelet = args.wavelet[0]

  if(args.singlechannel):
    numChannels = 1
  else:
    numChannels = 4

  for ch in range(numChannels):

    # CS    
    allplt_mag = [None]*len(args.wavelet)
    allplt_ang = [None]*len(args.wavelet)
    allplt_cmx = [None]*len(args.wavelet)
    # Linear
    allplt_mag_lin = [None]*len(args.wavelet)
    allplt_ang_lin = [None]*len(args.wavelet)
    allplt_cmx_lin = [None]*len(args.wavelet)

    for i,wavelet in enumerate(args.wavelet):

      mse_mag_cs, mse_ang_cs, mse_cmx_cs, perc_err_cs, mse_mag_lin, mse_ang_lin, mse_cmx_lin, perc_err_lin = get_error(ch,args.dir, args.maskdir, bl_noise, bl_uval, bl_utype, wavelet, bl_method, args.numsamples, args.usecompleximgs, args.usetrueimg, args.addlinearrec, args.usefluidmask, args.fluidmaskfile, args.usemultipatterns)
        
      allplt_mag[i] = mse_mag_cs
      allplt_ang[i] = mse_ang_cs
      allplt_cmx[i] = mse_cmx_cs

      if(args.addlinearrec):
        allplt_mag_lin[i] = mse_mag_lin
        allplt_ang_lin[i] = mse_ang_lin
        allplt_cmx_lin[i] = mse_cmx_lin

      if(args.usetrueimg):
        msg = 'tru'
        ylabel = 'MSE - w.r.t. true image'
      else:
        msg = 'avg'
        ylabel = 'MSE - w.r.t. average image'

      if(args.usecompleximgs):
        msg += '_cmx'
      else:
        msg += '_vel'

      # Store percent error
      comp = 'wavediff_k'+str(ch)+'_'+str(wavelet)+'_'+msg
      perc_dict[comp+'_cs'] = perc_err_cs
      perc_dict[comp+'_lin'] = perc_err_lin                

    # Generate Plots
    lgd = []
    for loopA in range(len(args.wavelet)):
      lgd.append(get_wavelet_string(args.wavelet[loopA]))

    # fig = generateViolinPlot(lgd,'Reconstruction method',allplt,allplt_lin)    
    fig = generateBarPlot(lgd,'Wavelet Frame',ylabel,
                          args.usecompleximgs,args.addlinearrec,
                          allplt_mag,allplt_ang,allplt_cmx,
                          allplt_mag_lin,allplt_ang_lin,allplt_cmx_lin,rotext=True)

    fname = args.outputdir + 'wavediff_' + msg + '_noise' + str(int(bl_noise*100)) + '_p'+str(int(bl_uval*100)) + '_' + bl_utype + '_k' + str(ch) + '.png'
    plt.savefig(fname,dpi=200)
    print("MSE Image saved: " + fname)
    plt.close(fig)    

def pltviolin(dir, recdir, patterndir, num_samples, use_complex, use_truth):
    #use_complex: compare against complex images or final recovered velocity images
    #use_truth: compare against true values or the average recovered images 
    for p in [0.25]:
        for type in ['vardengauss']:#'bernoulli', 'bpoisson']:
            for noise in [0.01, 0.05, 0.1, 0.3]:
                try: #plot comparison of undersampling % for CS and linear rec. images
                    plotnoisediff(dir, recdir, patterndir,  noise, p, type, num_samples, use_complex, use_truth, True)
                    plotnoisediff(dir, recdir, patterndir,  noise, p, type, num_samples, use_complex, use_truth, False)
                except Exception as e:
                    print(e)
                    print('Not found: recovered images with ', noise, 'noise', p, 'p', type, 'type')
                    continue
    for p in [0.25, 0.5, 0.75]:
        for type in ['vardengauss']:#'bernoulli', 'bpoisson']:
            for noise in [0.01]:
                try:
                    plotpdiff(dir, recdir, patterndir,  noise, p, type, num_samples, use_complex, use_truth, True)
                    plotpdiff(dir, recdir, patterndir,  noise, p, type, num_samples, use_complex, use_truth, False)
                except Exception as e:
                    print(e)
                    print("Not found: recovered images with ", noise, 'noise', p, 'p', type, 'type')
                    continue
    
# MAIN 
if __name__ == '__main__':

  # Init parser
  parser = argparse.ArgumentParser(description='Generate result images.')

  # Load Base Line Params
  # noise
  parser.add_argument('-n', '--noise',
                      action=None,
                      nargs='*',
                      const=None,
                      default=[0.1],
                      type=float,
                      choices=None,
                      required=False,
                      help='list of noise values',
                      metavar='',
                      dest='noise')
  # uval
  parser.add_argument('-u', '--uval',
                      action=None,
                      nargs='*',
                      const=None,
                      default=[0.75],
                      type=float,
                      choices=None,
                      required=False,
                      help='list of undersampling ratios',
                      metavar='',
                      dest='uval')
  # utype
  parser.add_argument('-t', '--utype',
                      action=None,
                      nargs='*',
                      const=None,
                      default=['vardengauss'],
                      type=str,
                      choices=None,
                      required=False,
                      help='list of random undersampling patterns',
                      metavar='',
                      dest='utype')
  # numsamples
  parser.add_argument('-s', '--numsamples',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=100,
                      type=int,
                      choices=None,
                      required=False,
                      help='number of repetitions',
                      metavar='',
                      dest='numsamples')
  # numpts
  parser.add_argument('-np', '--numpts',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=100,
                      type=int,
                      choices=None,
                      required=False,
                      help='number of points for computing the',
                      metavar='',
                      dest='numpts')

  # wavelet
  parser.add_argument('-w', '--wavelet',
                      action=None,
                      nargs='*',
                      const=None,
                      default=['haar'],
                      type=str,
                      choices=['haar','db8'],
                      required=False,
                      help='list of wavelet frames',
                      metavar='',
                      dest='wavelet')

  # method
  parser.add_argument('-m', '--method',
                      action=None,
                      nargs='*',
                      const=None,
                      default=['cs'],
                      type=str,
                      choices=None,
                      required=False,
                      help='list of reconstruction methods',
                      metavar='',
                      dest='method')
  # maindir
  parser.add_argument('-d', '--dir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='folder containing the file with the point locations for the correlation and with subfoldes with the reconstruction methods',
                      metavar='',
                      dest='dir')
  # maskdir
  parser.add_argument('-k', '--maskdir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='folder containing the undersampling masks',
                      metavar='',
                      dest='maskdir')  
  # outputdir
  parser.add_argument('-o', '--outputdir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='output folder for images',
                      metavar='',
                      dest='outputdir')
  # singlechannel
  parser.add_argument('--singlechannel',
                      action='store_true',
                      default=False,
                      required=False,
                      help='treat the image as single-channel, without velocity components',
                      dest='singlechannel')    
  # usecompleximgs
  parser.add_argument('--usecompleximgs',
                      action='store_true',
                      default=False,
                      required=False,
                      help='evaluate the MSE based on the reconstructed complex image. Default: use velocity components.',
                      dest='usecompleximgs')    

  # usefluidmask
  parser.add_argument('--usefluidmask',
                      action='store_true',
                      default=False,
                      required=False,
                      help='evaluate the MSE only within the fluid region. Default: use full image',
                      dest='usefluidmask')    

  # fluidmaskfile
  parser.add_argument('--fluidmaskfile',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='',
                      type=str,
                      choices=None,
                      required=False,
                      help='name of the npy file containing the binary mask',
                      metavar='',
                      dest='fluidmaskfile')  

  # addlinear reconstructions
  parser.add_argument('--addlinearrec',
                      action='store_true',
                      default=False,
                      required=False,
                      help='Plot linear reconstruction MSE.',
                      dest='addlinearrec')      
  # use true image
  parser.add_argument('--usetrueimg',
                      action='store_true',
                      default=False,
                      required=False,
                      help='use the true image to evaluate the MSE',
                      dest='usetrueimg')    

  # use unique undesampling pattern for each reconstruction (false) or change every time
  parser.add_argument('-um', '--usemultipatterns',
                      action='store_true',
                      default=False,
                      required=False,
                      help='generate a unique undersampling pattern for each noise realization',
                      dest='usemultipatterns')

  # Print Level
  parser.add_argument('-p', '--printlevel',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=0,
                      type=int,
                      choices=None,
                      required=False,
                      help='print level, 0 - no print, >0 increasingly more information ',
                      metavar='',
                      dest='printlevel')

  # String for perc_err files
  parser.add_argument('-ps', '--percstring',
                      action=None,
                      # nargs='*',
                      const=None,
                      default='',
                      type=str,
                      choices=None,
                      required=False,
                      help='string for percent error file',
                      metavar='',
                      dest='percstring')
  

  # Parse Commandline Arguments
  args = parser.parse_args()

  # Plot the parameter perturbations
  perc_dict = {}
  if(len(args.noise) > 1):
    plot_noisediff(args,perc_dict)
  if(len(args.uval) > 1):
    plot_pdiff(args,perc_dict)
  if(len(args.utype) > 1):
    plot_maskdiff(args,perc_dict)
  if(len(args.method) > 1):
    plot_methoddiff(args,perc_dict)
  if(len(args.method) > 1):
    plot_waveletdiff(args,perc_dict)

  # Save Dictionary to File
  perc_file = 'perc_dict_' + args.percstring + '.npy'
  print('Saving percent error dictionary to file: ',perc_file)
  np.save(perc_file, perc_dict)

  # Completed!
  if(args.printlevel > 0):
    print('Completed!!!')  



import vtk
from vtk.util.numpy_support import vtk_to_numpy
from vtk.util.numpy_support import numpy_to_vtk
import os
import sys
import numpy as np
import argparse
home = os.getenv('HOME')

def getInput(reader, filename):
    reader.SetFileName(filename)
    reader.Update()
    polyDataModel = reader.GetOutput()
    dims = polyDataModel.GetDimensions()
    data = polyDataModel.GetPointData()
    velocity = vtk_to_numpy(data.GetArray('velocity'))
    velocity = np.reshape(velocity, (dims[2], dims[1], dims[0],3))
    velocity = np.transpose(velocity, (3,2,1,0))
    concentration = vtk_to_numpy(data.GetScalars('concentration'))
    concentration = np.reshape(concentration, (dims[2], dims[1], dims[0],1))
    concentration = np.transpose(concentration, (3,2,1,0))
    return velocity,concentration

def getInputData(directory, vtkfile):
    reader = vtk.vtkRectilinearGridReader()#vtk.vtkStructuredPointsReader()
    reader.SetFileName(directory + vtkfile)
    reader.ReadAllVectorsOn()
    reader.ReadAllScalarsOn()
    reader.Update()
    velocity, concentration = getInput(reader,directory + vtkfile)
    data = np.concatenate((concentration,velocity),axis=0)
    return data

def saveInput(fromdir,vtkfile,todir,imagefile,sliceIndex,printlevel):
    inp = getInputData(fromdir, vtkfile)
    if(printlevel>0):
      print(inp.shape)
    inp = np.moveaxis(inp, 1+sliceIndex, 1)
    np.save(todir + imagefile + '.npy', inp)
    return inp

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Convert VTK rectilinear grid to compressed image.')
    
    parser.add_argument('-o', '--origindir',
                        action=None,
                        # nargs='+',
                        const=None,
                        default='/apps/pDat/samp256/',
                        type=str,
                        choices=None,
                        required=False,
                        help='origin VTK file folder',
                        metavar='',
                        dest='directory')

    parser.add_argument('-d', '--destdir',
                        action=None,
                        # nargs='+',
                        const=None,
                        default='/apps/undersampled/poiseuille/img/',
                        type=str,
                        choices=None,
                        required=False,
                        help='destination image folder',
                        metavar='',
                        dest='tosavedir')

    parser.add_argument('-s', '--sliceindex',
                        action=None,
                        # nargs='+',
                        const=None,
                        default=2,
                        type=str,
                        choices=None,
                        required=False,
                        help='slice index',
                        metavar='',
                        dest='sliceindex')

    parser.add_argument('-v', '--vtkname',
                        action=None,
                        # nargs='+',
                        const=None,
                        default='pout0_0.vtk',
                        type=str,
                        choices=None,
                        required=False,
                        help='name of the origin VTK file',
                        metavar='',
                        dest='vtkname')

    parser.add_argument('-f', '--imagename',
                        action=None,
                        # nargs='+',
                        const=None,
                        default='imgs_n1.npy',
                        type=str,
                        choices=None,
                        required=False,
                        help='name of the destination image file',
                        metavar='',
                        dest='imagename')

    parser.add_argument('-p', '--printlevel',
                        action=None,
                        # nargs='+',
                        const=None,
                        default=0,
                        type=str,
                        choices=None,
                        required=False,
                        help='print level, 0 - no print, >0 increasingly more information ',
                        metavar='',
                        dest='printlevel')

    # Parse Commandline Arguments
    args = parser.parse_args()

    # Convert
    saveInput(args.directory, 
              args.vtkname, 
              args.tosavedir, 
              args.imagename, 
              args.sliceindex,
              args.printlevel)

    if(args.printlevel > 0):
      print('Completed!!!')



import sys
import os
import math
import numpy as np
import numpy.fft as fft
import numpy.linalg as la
from mrina import generateSamplingMask, crop
import argparse
# import matplotlib.pyplot as plt

def getVenc(vel):
    mx = np.amax(np.fabs(vel))
    # less than 1/2: let it be equal to 1/3
    # Shouldn'd be max*1.5????
    # CHECK THIS !!!
    venc = mx*1.5
    # venc = mx*3.0
    return venc

def getKspace(sample, venc, magnitudes=None, referencephase=None):
    #return the fft of the complex images
    num_samples = sample.shape[0]
    yk = np.zeros(sample.shape,dtype=complex) #reference complex images for computing velocities
    if referencephase is None:
        print("Reference phase not found. Information loss may occur.")
        referencephase = np.zeros(sample[:,0].shape)
    if magnitudes is None:
        print("Magnitude of complex images not found. Information loss may occur.")
        magnitudes = np.ones(sample.shape)
    for n in range(num_samples):
        mag = sample[n,0]
        refphase = referencephase[n]
        vel = sample[n,1:4]
        compyk = np.zeros(vel.shape, dtype=complex)
        #2d fourier transform each slice for k space
        for k in range(sample.shape[2]): #number of 2d images in grid
            yk[n,0,k] = fft.fft2(mag[k]*np.exp(1j*refphase[k]),norm='ortho')
            for j in range(0,sample.shape[1]-1): #number of velocity components
                compyk[j,k] = magnitudes[n,j+1,k]*np.exp(1j*refphase[k])*np.exp(np.pi*1j*vel[j,k]/venc)
                yk[n,j+1,k] = fft.fft2(compyk[j,k],norm='ortho')
    return yk 

def undersample(kspace, mask):
    mag,vel = kspace
    for k in range(mag.shape[0]): #undersample differently for each sample
        mag[k,:,mask[k]] = 0
        vel[k,:,:,mask[k]] = 0
    return mag, vel

def get_noise(imsz, nrm, noise_percent, num_realizations, num_components=4):
    #noise param is a percentage of how much noise to be added
    noise = np.zeros((num_realizations, num_components, 1,) + imsz, dtype=complex)
    snr = np.zeros((num_realizations,num_components))

    for n in range(num_realizations):
        for j in range(num_components):
            # IS THERE A 2 MISSING HERE?? !!!
            avgnorm = nrm[j]/math.sqrt(np.prod(imsz))
            stdev = noise_percent * avgnorm
            if(stdev < 1.0e-12):           
              # There is no noise, so SNR is infinity
              snr[n,j] = np.infty
            else:
              snr[n,j] = (avgnorm/stdev)**2
            noise[n,j] = np.random.normal(scale=stdev, size=imsz) + 1j*np.random.normal(scale=stdev, size=imsz)
    return noise,snr

def add_noise(kspace, noise_percent, num_realizations, num_components=4):
    imsz = kspace.shape[3:]
    samples = np.zeros((num_realizations, num_components, 1,) + imsz, dtype=complex)
    nrm = [None]*num_components
    for v in range(kspace.shape[1]):
        nrm[v] = la.norm(kspace[0,v])
    print('K-space image component two-norms: ', nrm)
    noise, snr = get_noise(imsz, nrm, noise_percent, num_realizations , num_components)
    for n in range(num_realizations):
        for v in range(kspace.shape[1]):
            samples[n,v,0] = kspace[0,v,0] + noise[n,v]
    return samples, snr

def genSamples(fromdir,numRealizations,truefile,tosavedir,uType,uVal,uSeed,noisePercent,useMultiPatterns,printlevel):
    
    #p: percent not sampled, uType: bernoulli, vardengauss, halton, sliceIndex= 0,1,2 where to slice in grid
    #npydir: where to store numpy files, directory: where to retrieve vtk files, numSamples: # files to create
    
    # Get images
    if(printlevel>0):
        print('Loading Image...')
    inp = np.load(fromdir + truefile + '.npy')
    
    # Get velocity encoding
    venc = getVenc(inp[:,1:,:])
    
    # Save Velocity Encoding
    if(printlevel>0):
        print('Saving velocity encoding...')    
    # Always one velocity encoding for image set
    np.save(tosavedir + 'venc_n1.npy', venc) 

    # Transform image in the Fourier domain
    if(printlevel>0):
        print('Fourier transform in k-space...')    
    kspace = getKspace(inp, venc)
    
    # Generate undersampling mask
    if(printlevel>0):
        print('Generate and save sampling mask...')  
    
    undfile = tosavedir + 'undersamplpattern_p' + str(int(uVal*100)) + uType # + '_seed' + str(uSeed)       
    if useMultiPatterns:
      numPatterns = numRealizations
      undfile = undfile + "_n" + str(numPatterns)
    else:  
      numPatterns = 1

    mask = generateSamplingMask(kspace.shape[3:], uVal, saType=uType, num_patterns=numPatterns, seed=uSeed)
    np.save(undfile, mask)

    # Add noise to image
    if(printlevel>0):
        print('Add noise to image...')
    noisy,snr = add_noise(kspace, noisePercent, numRealizations)
    fourier_file = tosavedir + 'noisy_noise' + str(int(noisePercent*100)) + '_n' + str(numRealizations)
    np.save(fourier_file, noisy)
    
    # Save SNR
    if(printlevel>0):
        print('Save image SNR...')
    np.save(tosavedir + 'snr_noise' + str(int(noisePercent*100)) + '_n' + str(numRealizations), snr)

# --- MAIN FUNCTION
if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Generate noisy k-space measurements and undersampling masks.')
    # fromdir         = '\.'
    parser.add_argument('-f', '--fromdir',
                        action=None,
                        # nargs='+',
                        const=None,
                        default='./',
                        type=str,
                        choices=None,
                        required=False,
                        help='origin image location',
                        metavar='',
                        dest='fromdir')

    # numRealizations = 1
    parser.add_argument('-r', '--repetitions',
                        action=None,
                        # nargs='+',
                        const=None,
                        default=1,
                        type=int,
                        choices=None,
                        required=False,
                        help='number of k-space samples and mask to generate',
                        metavar='',
                        dest='repetitions')

    # truefile        = 'imgs_n1', 
    parser.add_argument('-o', '--origin',
                        action=None,
                        # nargs='+',
                        const=None,
                        default='imgs_n1',
                        type=str,
                        choices=None,
                        required=False,
                        help='name of the origin images to process',
                        metavar='',
                        dest='origin')

    # tosavedir       = None, 
    parser.add_argument('-d', '--dest',
                        action=None,
                        # nargs='+',
                        const=None,
                        default='./',
                        type=str,
                        choices=None,
                        required=False,
                        help='destination folder for sample generation',
                        metavar='',
                        dest='dest')

    # uType           = 'bernoulli',
    parser.add_argument('-u', '--utype',
                        action=None,
                        # nargs='+',
                        const=None,
                        default='bernoulli',
                        type=str,
                        choices=['bernoulli','vardentri','vardengauss','vardenexp','halton'],
                        required=False,
                        help='undersampling pattern type',
                        metavar='',
                        dest='uType')

    # uVal            = 0.5,
    parser.add_argument('-v', '--urate',
                        action=None,
                        # nargs='+',
                        const=None,
                        default=0.75,
                        type=float,
                        choices=None,
                        required=False,
                        help='undersampling rate',
                        metavar='',
                        dest='uVal')

    # uSeed           = 1234,
    parser.add_argument('-seed',
                        action=None,
                        # nargs='+',
                        const=None,
                        default=1234,
                        type=int,
                        choices=None,
                        required=False,
                        help='random generator seed',
                        metavar='',
                        dest='seed')

    # noisePercent    = 0.0
    parser.add_argument('-n', '--noisepercent',
                        action=None,
                        # nargs='+',
                        const=None,
                        default=0.0,
                        type=float,
                        choices=None,
                        required=False,
                        help='noise percent based the average two-norm of the k-space image',
                        metavar='',
                        dest='noisepercent')
    
    # useMultiPatterns = False
    parser.add_argument('-um', '--usemultipatterns',
                        action='store_true',
                        default=False,
                        required=False,
                        help='generate a unique undersamp. pattern for each noise realization',
                        dest='usemultipatterns')

    # printlevel = 0
    parser.add_argument('-p', '--printlevel',
                        action=None,
                        # nargs='+',
                        const=None,
                        default=0,
                        type=float,
                        choices=None,
                        required=False,
                        help='print level, 0 - no print, >0 increasingly more information ',
                        metavar='',
                        dest='printlevel')

  
    # Parse Commandline Arguments
    args = parser.parse_args()
    
    # Generate Samples
    genSamples(args.fromdir,
               args.repetitions,
               args.origin,
               args.dest,
               args.uType,
               args.uVal,
               args.seed,
               args.noisepercent,
               args.usemultipatterns,
               args.printlevel)

    if(args.printlevel > 0):
      print('Completed!!!')



# %% 
import numpy as np
import numpy.linalg as la
import time
import scipy.optimize as sciopt

# %%
# Soft-thresholding
def sft(x, t):
    return x * (1.0 - np.minimum(t / np.maximum(np.abs(x), 1E-32), 1.0))

# Projection onto the l1-ball
def project_l1_ball(x, t):
    xL1Nrm  = la.norm(x.ravel(), 1)
    if xL1Nrm > t:
        # If the input is outside the l1-ball we project.
        nx = x.size;
        # Sort magnitudes in decreasing order
        mx = np.flip(np.sort(np.absolute(x.ravel()), kind='mergesort'))
        # Obtain unique values (this is needed for some borderline cases)
        umx = np.flip(np.unique(mx))
        if t <= umx[0] - umx[1]:
            # If the radius of the l1-ball is small and the input is far away
            # and the shrinkage is large.
            vp = umx[0]
            sroot = vp - t
            return sft(x, sroot)
        else:
            # This is the average case
            cmx = np.cumsum(mx)
            smx = cmx - np.array(range(1, nx+1)) * mx
        if smx[nx - 1] < t:
            # This condition handles some cases when the input is close to
            # the boundary and the shrinkage is small.
            sroot = (xL1Nrm - t) / nx
            return sft(x, sroot)
        idp, = np.where( smx > t )
        idp = idp[0];
        sroot = (cmx[idp - 1] - t) / idp
        return sft(x, sroot)
    # If the input is within the l1-ball we do nothing.
    return x

# Minimize Sum-of-Squares
#   min_x || Ax - y ||_2^2
def MinimizeSumOfSquares(y, A, xinit=None, 
                            maxItns=1E4, 
                            dxAbsTol=1E-6, dxRelTol=1E-9,
                            gradAbsTol=1E-5, gradRelTol=1E-8,
                            disp=False, printEvery=100):
    # Find Lipschitz constant of the gradient
    L = 2.05 * A.norm() ** 2
    # Initialize variables
    if xinit is None:
        x = np.zeros(A.wavShape, dtype=np.complex)
    else:
        x = xinit
    xNrm = la.norm(x.ravel(), 2)
    dxNrm = np.inf
    itn = 0
    fx = la.norm((y - A.eval(x)).ravel(), 2) ** 2
    # Optimization loop
    stop = False
    if disp:
        print('[MinimizeSumOfSquares]')
        print(' Initial objective:   {:1.6e}'.format(fx))
    while not stop: 
        itn = itn + 1
        xp = x - (2.0/L) * A.adjoint(A.eval(x) - y)
        fxp = la.norm((y - A.eval(xp)).ravel(), 2) ** 2
        dxNrm = la.norm((xp - x).ravel(), 2)
        gradNrm = L * dxNrm
        dxNrmTol = np.max([ dxAbsTol, 
                            dxRelTol * xNrm, 
                            L * gradAbsTol, 
                            L * gradRelTol * xNrm ])
        if dxNrm < dxNrmTol or itn > maxItns:
            stop = True
        x = xp
        fx = fxp
        xNrm = la.norm(x.ravel(), 2)
        if disp and (itn == 1 or np.mod(itn, printEvery) == 0 or stop):
            print(' {:04d} | obj: {:1.3E} | gradNrm: {:1.3E}'.format(itn, fx, gradNrm))
    return x, fx

# Minimize Sum-of-Squares subject to l1-norm
#   min_x || Ax - y ||_2^2 s.t. ||x||_1 <= t
def MinimizeSumOfSquaresL1Ball(t, y, A, xinit=None, L=None,
                                maxItns=1E4, 
                                dxAbsTol=1E-4, dxRelTol=1E-6,
                                dpAbsTol=1E-5, dpRelTol=1E-8,
                                disp=False, printEvery=10,
                                restart=True):
    # Project initial iterate on the l1-ball
    if xinit is None:
        x = np.zeros(A.inShape, dtype=np.complex)
    else:
        x = project_l1_ball(xinit, t)
    if L is None:
        L = 2.05 * A.norm() ** 2
    rx = A.eval(x) - y
    fx = la.norm(rx.ravel()) ** 2
    gx = 2.0 * A.adjoint(rx)
    rpx = x - project_l1_ball(x - gx/L, t)
    rpxNrm = la.norm(rpx.ravel())
    # Initialize variables
    xNrm = la.norm(x.ravel())
    dxNrm = np.inf
    dfx = -np.inf
    itn = 0
    s = 1
    z = x
    # Optimization loop
    stop = False
    if disp:
        print('[PGD-SoS-L1Ball]')
        print(' Initial objective:   {:1.6e}'.format(fx))
        print(' Initial residual:    {:1.6e}'.format(rpxNrm))
    while not stop:
        itn = itn + 1
        gz = 2.0 * A.adjoint(A.eval(z, 1) - y)
        xp = project_l1_ball(z - gz/L, t)
        sp = 0.5 * (1.0 + np.sqrt(1.0 + 4.0 * s ** 2))
        zp = xp + ((s - 1.0)/sp) * (xp - x)

        # Update
        rxp = A.eval(xp) - y
        fxp = la.norm(rxp.ravel()) ** 2

        dfx = fxp - fx
        dxNrm = la.norm((xp - x).ravel())

        if restart and dfx > 1E-12:
            z = x
            s = 1.0
            dfx = np.inf
            dxNrm = np.inf
        else:
            x = xp
            gx = 2.0 * A.adjoint(rxp)
            rpx = x - project_l1_ball(x - gx/L, t)
            rpxNrm = la.norm(rpx.ravel())

            z = zp
            fx = fxp
            s = sp

        xNrm = la.norm(x.ravel())
        dxTol = np.max([ dxAbsTol, dxRelTol * xNrm ])
        rpxTol = np.max([ dpAbsTol, dpRelTol * xNrm ]) 
        if (dxNrm < dxTol and rpxNrm < rpxTol) or itn > maxItns:
            stop = True
        if disp and (itn == 1 or np.mod(itn, printEvery) == 0 or stop):
            print(' {:04d} | obj: {:1.3E} | dfx: {:+1.3E} | dxNrm: {:1.3E} | rpxNrm: {:5.3E}'.format(itn, fx, dfx, dxNrm, rpxNrm))
    return x, fx

# Minimize Sum-of-Squares
#   min_x || Ax - y ||_2^2 + t ||x||_1
def MinimizeBPDN(t, y, A, xinit=None, L=None,
                    maxItns=1E4, 
                    dxAbsTol=1E-4, dxRelTol=1E-6,
                    dpAbsTol=1E-5, dpRelTol=1E-8,
                    disp=False, printEvery=10,
                    restart=True):
    # Project initial iterate on the l1-ball
    if xinit is None:
        x = np.zeros(A.inShape, dtype=np.complex)
    else:
        x = project_l1_ball(xinit, t)
    z = x
    if L is None:
        L = 2.05 * A.norm() ** 2
    rx = A.eval(x) - y
    fx = la.norm(rx.ravel()) ** 2 + t * la.norm(x.ravel(), 1)
    gx = 2.0 * A.adjoint(rx)
    rpx = x - sft(x - gx, t)
    rpxNrm = la.norm(rpx.ravel())
    # Initialize variables
    xNrm = la.norm(x.ravel())
    dxNrm = np.inf
    dfx = -np.inf
    itn = 0
    s = 1.0
    # Optimization loop
    stop = False
    if disp:
        print('[PGD-BPDN]')
        print(' Initial objective:   {:1.6e}'.format(fx))
        print(' Initial residual:    {:1.6e}'.format(rpxNrm))
    while not stop:
        itn = itn + 1
        gz = 2.0 * A.adjoint(A.eval(z, 1) - y)
        xp = sft(z - gz/L, t/L)
        sp = 0.5 * (1.0 + np.sqrt(1.0 + 4.0 * s ** 2))
        zp = xp + ((s - 1.0)/sp) * (xp - x)

        # Update
        rxp = A.eval(xp) - y
        fxp = la.norm(rxp.ravel()) ** 2 + t * la.norm(xp.ravel(), 1)

        dfx = fxp - fx
        dxNrm = la.norm((xp - x).ravel())

        if restart and dfx > 1E-12:
            z = x
            s = 1.0
            dfx = np.inf
            dxNrm = np.inf
        else:
            x = xp
            gx = 2.0 * A.adjoint(rxp)
            rpx = x - sft(x - gx, t)
            rpxNrm = la.norm(rpx.ravel())

            z = zp
            fx = fxp
            s = sp

        xNrm = la.norm(x.ravel())
        dxTol = np.max([ dxAbsTol, dxRelTol * xNrm ])
        rpxTol = np.max([ dpAbsTol, dpRelTol * xNrm ]) 
        if (dxNrm < dxTol and rpxNrm < rpxTol) or itn > maxItns:
            stop = True

        if disp and (itn == 1 or np.mod(itn, printEvery) == 0 or stop):
            print(' {:04d} | obj: {:1.3E} | dfx: {:+1.3E} | dxNrm: {:1.3E} | rpxNrm: {:5.3E}'.format(itn, fx, dfx, dxNrm, rpxNrm))
    return x, fx

class RootSolverL1NormNoisy():
    def __init__(self, eta, y, A,
                    maxItns=1E4, 
                    dxAbsTol=1E-4, dxRelTol=1E-6,
                    dpAbsTol=1E-5, dpRelTol=1E-8,
                    disp=False, printEvery=10,
                    restart=True,
                    method='SoS-L1Ball', disp_method=False):
        # Problem parameters
        self.eta = eta
        self.y = y
        self.A = A
        self.disp_method = disp_method

        self.ANrm = A.norm()
        self.L = 2.05 * self.ANrm ** 2

        # For bracketing
        self._itn = 0
        self._t = [ None, None ]
        self._ft = [ None, None ]
        self._xinit = [ None, None ]

        _xopt, _ = MinimizeSumOfSquares(y, A, 
                                            maxItns=maxItns, 
                                            dxAbsTol=dxAbsTol, dxRelTol=dxRelTol,
                                            gradAbsTol=dpAbsTol, gradRelTol=dpRelTol,
                                            disp=False)
        _ropt = A.eval(_xopt) - y
        if method == 'SoS-L1Ball':
            # Lower bound
            self._t[0] = 0.0
            self._ft[0] = la.norm(y.ravel()) - eta
            self._xinit[0] = np.zeros(A.inShape, dtype=np.complex)
            # Upper bound
            self._t[1] = la.norm(_xopt.ravel(), 1)
            self._ft[1] = la.norm(_ropt.ravel(), 2) - eta
            self._xinit[1] = _xopt
        elif method == 'BPDN':
            # Lower bound
            self._t[0] = 0.0
            self._ft[0] = -eta
            self._xinit[0] = _xopt
            # Upper bound
            _z = A.adjoint(y)
            self._t[1] = la.norm(_z.ravel(), np.inf)
            self._ft[1] = la.norm(y.ravel(), 2) - eta
            self._xinit[1] = np.zeros(A.inShape, dtype=np.complex)
        else:
            raise ValueError('Method can be either SoS-L1Ball or BPDN.')
        self.method = method
        # Solver options
        self.restart = restart
        # Solver parameters
        self.maxItns = maxItns
        self.dxAbsTol = dxAbsTol
        self.dxRelTol = dxRelTol
        self.dpAbsTol = dpAbsTol
        self.dpRelTol = dpRelTol
        # Display options
        self.disp = disp
        self.printEvery = printEvery

    def solve(self, t):
        self._itn = self._itn + 1
        if self.disp:
            print(' {:02d} | t* = {:1.4E} in [{:1.4E}, {:1.4E}] | f: [{:+1.4E}, {:+1.4E}]'.format(self._itn, t, self._t[0], self._t[1], self._ft[0], self._ft[1]))

        if np.abs(t - self._t[0]) < np.abs(t - self._t[1]):
            tNear = 0
        else:
            tNear = 1

        theta = np.abs(t - self._t[0]) / (np.abs(t - self._t[0]) + np.abs(t - self._t[1]))
        xinit = theta * self._xinit[0] + (1.0 - theta) * self._xinit[1]

        if self.method == 'SoS-L1Ball':
            xopt, _ = MinimizeSumOfSquaresL1Ball(t, self.y, self.A, xinit=xinit, L=self.L,
                                                    maxItns=self.maxItns, 
                                                    dxAbsTol=self.dxAbsTol, dxRelTol=self.dxRelTol,
                                                    dpAbsTol=self.dpAbsTol, dpRelTol=self.dpRelTol,
                                                    disp=self.disp_method, printEvery=self.printEvery,
                                                    restart=self.restart)
        elif self.method == 'BPDN':
            xopt, _ = MinimizeBPDN(t, self.y, self.A, xinit=xinit, L=self.L,
                                    maxItns=self.maxItns, 
                                    dxAbsTol=self.dxAbsTol, dxRelTol=self.dxRelTol,
                                    dpAbsTol=self.dpAbsTol, dpRelTol=self.dpRelTol,
                                    disp=self.disp_method, printEvery=self.printEvery,
                                    restart=self.restart)

        ft = la.norm((self.A.eval(xopt) - self.y).ravel(), 2) - self.eta

        if t > self._t[1]:
            self._t[0] = self._t[1]
            self._ft[0] = self._ft[1]
            self._xinit[0] = self._xinit[1]

            self._t[1] = t
            self._ft[1] = ft
            self._xinit[1] = xopt
        elif t < self._t[0]:
            self._t[1] = self._t[0]
            self._ft[1] = self._ft[0]
            self._xinit[1] = self._xinit[0]

            self._t[0] = t
            self._ft[0] = ft
            self._xinit[0] = xopt
        else:
            self._t[tNear] = t
            self._ft[tNear] = ft
            self._xinit[tNear] = xopt

        return ft

def RecoveryL1NormNoisy(eta, y, A, 
                    maxItns=1E4, 
                    dxAbsTol=1E-4, dxRelTol=1E-6,
                    dpAbsTol=1E-5, dpRelTol=1E-8,
                    disp=False, printEvery=100,
                    method='BPDN', disp_method=False):

    # Create solver object
    T = RootSolverL1NormNoisy(eta, y, A,
                            maxItns=maxItns, 
                            dxAbsTol=dxAbsTol, dxRelTol=dxRelTol,
                            dpAbsTol=dpAbsTol, dpRelTol=dpRelTol,
                            disp=disp, printEvery=printEvery,
                            restart=True, method=method, disp_method=disp_method)
    if disp:
        print('[RECOVERY: L1-NOISY]' )
        print(' Method:          {:s}'.format(method) )
        print(' Bound:           {:1.4E}'.format(eta) )
        print(' Initial values:')
        print('     t           [{:1.4E},   {:1.4E}]'.format(T._t[0], T._t[1]))
        print('     f(t)        [{:+1.4E}, {:+1.4E}]'.format(T._ft[0], T._ft[1]))
    # Root finding via TOMS748
    tStart = time.time()
    tMin = T._t[0]
    tMax = T._t[1]
    topt, rnfo = sciopt.toms748(lambda t: T.solve(t), tMin, tMax, k=1, xtol=1E-4, full_output=disp, disp=disp)
    if np.abs(topt - T._t[0]) < np.abs(topt - T._t[1]):
        xopt = T._xinit[0]
    else:
        xopt = T._xinit[1]
    tStop = time.time() - tStart
    if disp:
        print('  Summary');
        print('      Iterations:          {:d}'.format(rnfo.iterations) )
        print('      Function calls:      {:d}'.format(rnfo.function_calls) )
        print('      Optimal value:       {:5.3E}'.format(topt) )
        print('      Elapsed time:        {:8.4f} seconds'.format(tStop) )
    return xopt, la.norm(xopt.ravel(), 1)

# %%
if __name__ == '__main__':
    # Import
    from maps import OperatorWaveletToFourier, OperatorWaveletToFourierX4
    
    do_single_image = True
    do_x4_images = False

    print('[TESTING SOLVER ROUTINES]')
    # Set up map
    imShape = (128, 128)
    # *** EXPERIMENTS FOR SINGLE IMAGE AND SUBSAMPLING - L1-NORM
    if do_single_image:
        samplingSet = np.where(np.random.uniform(size=imShape) < 0.75, True, False)
        A = OperatorWaveletToFourier(imShape, samplingSet=samplingSet, waveletName='db4')
        xinit = np.random.normal(size=A.inShape) + 1j * np.random.normal(size=A.inShape)
        y = A.eval(xinit)
        print('[SINGLE IMAGE - RANDOM UNDERSAMPLING - L1-NORM RECOVERY]')
        print('    Input shape:    {:d} x {:d}'.format(A.inShape[0], A.inShape[1]))
        print('    Output shape:   {:d} x {:d}'.format(A.outShape[0], 1))
        print('    Operator norm:  {:1.3e}'.format(A.norm()))
        # Tests
        print('\n[TEST: MinimizeSumOfSquares]')
        xsol, fmin = MinimizeSumOfSquares(y, A, disp=True)

        print('\n[TEST: MinimizeSumOfSquaresL1Ball]')
        print('  ---> with no restart...')
        xsol, fmin = MinimizeSumOfSquaresL1Ball(0.25, y, A, disp=True, restart=False)
        print('  ---> with restart...')
        _xsol, _fmin = MinimizeSumOfSquaresL1Ball(0.25, y, A, disp=True, restart=True)
        print('  ---> Difference:     {:1.5e}'.format(la.norm((xsol - _xsol).ravel())))

        print('\n[TEST: RecoveryL1Noisy - SoS-L1Ball]')
        xsol, fmin = RecoveryL1NormNoisy(10, y, A, maxItns=1E4, disp=True, printEvery=100, method='SoS-L1Ball')

        print('\n[TEST: RecoveryL1Noisy - BPDN]')
        _xsol, _fmin = RecoveryL1NormNoisy(10, y, A, maxItns=1E4, disp=True, printEvery=100, method='BPDN')
        print('  ---> Difference:')
        print('         Objective:    {:1.5e}'.format(np.abs(fmin - _fmin)))
        print('         Solution:     {:1.5e}'.format(la.norm((xsol - _xsol).ravel())))

    # *** EXPERIMENTS FOR MULTIPLE IMAGES AND SUBSAMPLING - L1-NORM
    if do_x4_images:
        samplingSet = np.where(np.random.uniform(size=imShape + (4,)) < 0.75, True, False)
        A = OperatorWaveletToFourierX4(imShape, samplingSet=samplingSet, waveletName='db4')
        xinit = np.random.normal(size=A.inShape) + 1j * np.random.normal(size=A.inShape)
        y = A.eval(xinit)
        print('[4 IMAGES - RANDOM UNDERSAMPLING - L1-NORM RECOVERY]')
        print('    Input shape:    {:d} x {:d} x {:d}'.format(A.inShape[0], A.inShape[1], A.inShape[2]))
        print('    Output shape:   {:d} x {:d}'.format(A.outShape[0], 1))
        print('    Operator norm:  {:1.3e}'.format(A.norm()))
        # Tests
        print('\n[TEST: MinimizeSumOfSquares]')
        xsol, fmin = MinimizeSumOfSquares(y, A, disp=True)

        print('\n[TEST: MinimizeSumOfSquaresL1Ball]')
        print('  ---> with no restart...')
        xsol, fmin = MinimizeSumOfSquaresL1Ball(0.25, y, A, disp=True, restart=False)
        print('  ---> with restart...')
        _xsol, _fmin = MinimizeSumOfSquaresL1Ball(0.25, y, A, disp=True, restart=True)
        print('  ---> Difference:     {:1.5e}'.format(la.norm((xsol - _xsol).ravel())))

        print('\n[TEST: RecoveryL1Noisy - SoS-L1Ball]')
        xsol, fmin = RecoveryL1NormNoisy(10, y, A, maxItns=1E4, disp=True, printEvery=100, method='SoS-L1Ball')

        print('\n[TEST: RecoveryL1Noisy - BPDN]')
        _xsol, _fmin = RecoveryL1NormNoisy(10, y, A, maxItns=1E4, disp=True, printEvery=100, method='BPDN')
        print('  ---> Difference:')
        print('         Objective:    {:1.5e}'.format(np.abs(fmin - _fmin)))
        print('         Solution:     {:1.5e}'.format(la.norm((xsol - _xsol).ravel())))
# %%
